# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import StrEnum

from faebryk.core.module import Module


class has_designator_prefix(Module.TraitT.decless()):
    class Prefix(StrEnum):
        A = "A"
        """Separable assembly or sub-assembly (e.g. printed circuit assembly)"""

        AT = "AT"
        """Attenuator or isolator"""

        BR = "BR"
        """
        Bridge rectifier (four diodes in a package) > often changed to "D" for diode
        """

        B = "B"
        """Often shortened to "B" for Battery or battery holder"""

        BT = "BT"
        """Battery or battery holder > often shortened to "B" """

        BAT = "BAT"
        """Battery or battery holder > often shortened to "B" """

        C = "C"
        """Capacitor"""

        CB = "CB"
        """Circuit breaker"""

        CN = "CN"
        """Capacitor network > may be simplified to "C" for capacitor"""

        D = "D"
        """
        Diode (all types, including LED), thyristor > "D" is preferred for various types
        of diodes
        """

        CR = "CR"
        """
        Diode (all types, including LED), thyristor > "D" is preferred for various types
        of diodes
        """

        DL = "DL"
        """Delay line"""

        DN = "DN"
        """Diode network > may be simplified to "D" for diode"""

        DS = "DS"
        """Display, general light source, lamp, signal light"""

        F = "F"
        """Fuse"""

        FB = "FB"
        """Ferrite bead > sometimes changed to "L" for inductor, though "E" was used in
        the currently inactive standard IEEE 315 (see Clause 22.4)"""

        L = "L"
        """Inductor or coil or ferrite bead > sometimes changed from "FB" for ferrite
        bead"""

        FD = "FD"
        """Fiducial"""

        FL = "FL"
        """Filter"""

        G = "G"
        """Generator or oscillator"""

        GDT = "GDT"
        """Gas discharge tube"""

        OSC = "OSC"
        """Generator or oscillator"""

        GL = "GL"
        """Graphical logo"""

        GN = "GN"
        """General network"""

        H = "H"
        """Hardware, e.g., screws, nuts, washers, also used for drilled holes >
        sometimes hardware is expanded to "HW" """

        HW = "HW"
        """Expanded form of "H" for hardware"""

        HY = "HY"
        """Circulator or directional coupler"""

        IR = "IR"
        """Infrared diode > often changed to "D" for diode"""

        J = "J"
        """Jack (least-movable connector of a connector pair), jack connector (connector
        may have "male" pin contacts and/or "female" socket contacts) > all types of
        connectors, including pin headers."""

        JP = "JP"
        """Jumper (link)"""

        K = "K"
        """Relay or contactor"""

        LD = "LD"
        """LED > often changed to "D" for diode"""

        LED = "LED"
        """LED > often changed to "D" for diode"""

        LS = "LS"
        """Loudspeaker or buzzer"""

        SPK = "SPK"
        """Loudspeaker or buzzer"""

        M = "M"
        """Motor"""

        MK = "MK"
        """Microphone"""

        MIC = "MIC"
        """Microphone"""

        MOD = "MOD"
        """Module"""

        MP = "MP"
        """Mechanical part (including screws and fasteners)"""

        OP = "OP"
        """Opto-isolator > often changed to "U" for IC"""

        U = "U"
        """Shorter form of "U" (unit) preferred for Integrated Circuit instead of "IC"
        """

        IC = "IC"
        """Integrated circuit (IC) > often shortened to "U" """

        P = "P"
        """Plug (most-movable connector of a connector pair), plug connector (connector
        may have "male" pin contacts and/or "female" socket contacts)"""

        PS = "PS"
        """Power supply"""

        Q = "Q"
        """Transistor (all types)"""

        R = "R"
        """Resistor"""

        RN = "RN"
        """Resistor network > sometimes simplified to "R" for resistor, or "N" for
        network"""

        N = "N"
        """Often used for networks, sometimes simplified from resistor network "RN" """

        RT = "RT"
        """Thermistor > sometimes simplified to "R" for resistor"""

        RV = "RV"
        """Varistor, variable resistor"""

        S = "S"
        """Switch (all types, including buttons) > sometimes "SW" is erroneously used"""

        SW = "SW"
        """Sometimes erroneously used for Switch instead of "S" """

        SA = "SA"
        """Spark arrester"""

        T = "T"
        """Transistor > often changed to "Q", but sometimes "T" is used for bipolar
        transistors and "Q" for FETs."""

        TC = "TC"
        """Thermocouple"""

        TP = "TP"
        """Test point"""

        TR = "TR"
        """Transformer > sometimes changed to "L" for inductor"""

        TUN = "TUN"
        """Tuner"""

        V = "V"
        """Vacuum tube"""

        VR = "VR"
        """Voltage regulator (voltage reference), or variable resistor (potentiometer /
        trimmer / rheostat) > voltage regulators are often "U" for IC, pots and trimmers
        often "R" for resistor"""

        X = "X"
        """Socket connector for another item not P or J, paired with the letter symbol
        for that item (XV for vacuum tube socket, XF for fuse holder, XA for printed
        circuit assembly connector, XU for integrated circuit connector, XDS for light
        socket, etc.)"""

        XTAL = "XTAL"
        """Crystal, ceramic resonator, powered oscillator"""

        Y = "Y"
        """Crystal, ceramic resonator, powered oscillator"""

        ZD = "ZD"
        """Zener diode > often changed to "D" for diode"""

    def __init__(self, prefix: str | Prefix) -> None:
        super().__init__()
        self.prefix = prefix

    def get_prefix(self) -> str:
        return self.prefix



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_single_connection_impl(F.has_single_connection.impl()):
    def get_connection(self):
        conns = self.obj.connections
        assert len(conns) == 1
        return conns[0]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class Crystal(Module):
    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    gnd: F.Electrical
    unnamed = L.list_field(2, F.Electrical)

    # ----------------------------------------
    #               parameters
    # ----------------------------------------
    frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(32.768 * P.kHz, 100 * P.MHz),
        tolerance_guess=50 * P.ppm,
    )

    frequency_tolerance = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(10 * P.ppm, 100 * P.ppm),
        tolerance_guess=10 * P.percent,
    )

    frequency_temperature_tolerance = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(1 * P.ppm, 50 * P.ppm),
        tolerance_guess=10 * P.percent,
    )

    frequency_ageing = L.p_field(
        units=P.ppm,
        likely_constrained=True,
        soft_set=L.Range(1 * P.ppm, 10 * P.ppm),
        tolerance_guess=20 * P.percent,
    )

    equivalent_series_resistance = L.p_field(
        units=P.Ω,
        likely_constrained=True,
        soft_set=L.Range(10 * P.Ω, 200 * P.Ω),
        tolerance_guess=10 * P.percent,
    )

    shunt_capacitance = L.p_field(
        units=P.F,
        likely_constrained=True,
        soft_set=L.Range(1 * P.pF, 10 * P.pF),
        tolerance_guess=20 * P.percent,
    )

    load_capacitance = L.p_field(
        units=P.F,
        likely_constrained=True,
        soft_set=L.Range(8 * P.pF, 30 * P.pF),
        tolerance_guess=10 * P.percent,
    )

    # ----------------------------------------
    #                traits
    # ----------------------------------------
    designator = L.f_field(F.has_designator_prefix)(F.has_designator_prefix.Prefix.XTAL)
    footprint: F.can_attach_to_footprint_symmetrically

    # ----------------------------------------
    #                connections
    # ----------------------------------------



from unittest.mock import Mock

import pytest

from faebryk.core.node import Node
from faebryk.exporters.netlist.graph import (
    _conflicts,
    _NetName,
)
from faebryk.library import Net as F


@pytest.mark.parametrize(
    "base_name,prefix,suffix,expected",
    [
        ("test", None, None, "test"),
        ("test", "prefix1-prefix2", None, "prefix1-prefix2-test"),
        ("test", None, 1, "test-1"),
        ("test", "prefix", 2, "prefix-test-2"),
        (None, "prefix", 1, "prefix-net-1"),
        (None, None, None, "net"),
    ],
)
def test_net_name(base_name, prefix, suffix, expected):
    net = _NetName(base_name=base_name, prefix=prefix, suffix=suffix)
    assert net.name == expected


@pytest.fixture
def node_hierarchy():
    """Creates a mock node hierarchy for testing."""
    root = Mock(spec=Node)
    child1 = Mock(spec=Node)
    child2 = Mock(spec=Node)
    grandchild1 = Mock(spec=Node)
    grandchild2 = Mock(spec=Node)

    # Setup hierarchies
    root.get_hierarchy = lambda: [(root, "root")]
    child1.get_hierarchy = lambda: [(root, "root"), (child1, "child1")]
    child2.get_hierarchy = lambda: [(root, "root"), (child2, "child2")]
    grandchild1.get_hierarchy = lambda: [
        (root, "root"),
        (child1, "child1"),
        (grandchild1, "grandchild1"),
    ]
    grandchild2.get_hierarchy = lambda: [
        (root, "root"),
        (child2, "child2"),
        (grandchild2, "grandchild2"),
    ]

    return type(
        "NodeHierarchy",
        (),
        {
            "root": root,
            "child1": child1,
            "child2": child2,
            "grandchild1": grandchild1,
            "grandchild2": grandchild2,
        },
    )


@pytest.fixture
def mock_nets():
    """Creates mock nets for conflict testing."""
    return [Mock(spec=F.Net) for _ in range(4)]


def test_conflicts_no_conflicts(mock_nets):
    names = {
        mock_nets[0]: _NetName(base_name="test1"),
        mock_nets[1]: _NetName(base_name="test2"),
    }
    assert list(_conflicts(names)) == []


def test_conflicts_with_single_conflict(mock_nets):
    names = {
        mock_nets[0]: _NetName(base_name="test"),
        mock_nets[1]: _NetName(base_name="test"),
    }
    conflicts = list(_conflicts(names))
    assert len(conflicts) == 1
    assert len(conflicts[0]) == 2
    assert all(net in conflicts[0] for net in [mock_nets[0], mock_nets[1]])


def test_conflicts_with_multiple_conflicts(mock_nets):
    names = {
        mock_nets[0]: _NetName(base_name="test1"),
        mock_nets[1]: _NetName(base_name="test1"),
        mock_nets[2]: _NetName(base_name="test2"),
        mock_nets[3]: _NetName(base_name="test2"),
    }
    conflicts = list(_conflicts(names))
    assert len(conflicts) == 2
    assert all(len(conflict) == 2 for conflict in conflicts)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P
from test.common.resources.fabll_modules.USB2_0_ESD_Protection import (
    USB2_0_ESD_Protection,
)
from test.common.resources.fabll_modules.USB_Type_C_Receptacle_14_pin_Vertical import (
    USB_Type_C_Receptacle_14_pin_Vertical,
)


# TODO bad module
class USB_C_PSU_Vertical(Module):
    # interfaces
    power_out: F.ElectricPower
    usb: F.USB2_0

    # components

    usb_connector: USB_Type_C_Receptacle_14_pin_Vertical  # TODO: make generic
    configuration_resistors = L.list_field(2, F.Resistor)
    gnd_resistor: F.Resistor
    gnd_capacitor: F.Capacitor
    esd: USB2_0_ESD_Protection

    def __preinit__(self):
        self.gnd_capacitor.capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.05)
        )
        self.gnd_capacitor.max_voltage.constrain_subset(
            L.Range.from_center_rel(16 * P.V, 0.05)
        )
        self.gnd_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1 * P.Mohm, 0.05)
        )
        for res in self.configuration_resistors:
            res.resistance.constrain_subset(L.Range.from_center_rel(5.1 * P.kohm, 0.05))

        vbus = self.usb_connector.vbus.fused(self)
        vbus_fuse = self.get_first_child_of_type(F.Fuse)
        vbus_fuse.fuse_type.constrain_subset(F.Fuse.FuseType.RESETTABLE)
        vbus_fuse.trip_current.constrain_subset(L.Range.from_center_rel(1 * P.A, 0.05))

        # alliases
        gnd = vbus.lv

        # connect usb and esd
        self.usb_connector.usb.connect_via(self.esd, self.usb)
        vbus.connect(self.power_out)

        # configure as ufp with 5V@max3A
        self.usb_connector.cc1.connect_via(self.configuration_resistors[0], gnd)
        self.usb_connector.cc2.connect_via(self.configuration_resistors[1], gnd)

        # EMI shielding
        self.usb_connector.shield.connect_via(self.gnd_resistor, gnd)
        self.usb_connector.shield.connect_via(self.gnd_capacitor, gnd)

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(self.power_out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.util import (  # noqa: F401
    KeyErrorAmbiguous,
    KeyErrorNotFound,
    find,
    not_none,
)

logger = logging.getLogger(__name__)


class Pinmux(Module):
    """
    Generic Pinmux Base Class
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------

        self._function_matrix = self._get_matrix()
        self._ios = self._get_ios()
        self.configured: dict[F.Electrical, F.Electrical] = {}

    def _get_ios(self) -> list[F.Electrical]:
        raise L.AbstractclassError()

    def _get_matrix(self) -> dict[F.Electrical, list[F.Electrical | None]]:
        raise L.AbstractclassError()

    def set_function(self, pin: int | F.Electrical, function: int | F.Electrical):
        if isinstance(pin, int):
            pin = self._ios[pin]
        if isinstance(function, int):
            function = not_none(self._function_matrix[pin][function])

        if pin in self.configured:
            if self.configured[pin].is_connected_to(function):
                return
            raise ValueError(f"Pin {pin} already configured")
        self.configured[pin] = function
        pin.connect(function)

    def enable(
        self, bus: ModuleInterface, pins: list[int | F.Electrical] | None = None
    ):
        signals = [
            logic.line
            for logic in bus.get_children(
                direct_only=False, types=F.ElectricLogic, include_root=True
            )
        ]

        candidates = {
            signal: [
                pin
                for pin, functions in self._function_matrix.items()
                if signal in functions
            ]
            for signal in signals
        }

        if pins is not None:
            pins = [self._ios[pin] if isinstance(pin, int) else pin for pin in pins]

        for signal, pin_candidates in candidates.items():
            if not pin_candidates:
                # TODO maybe allow it
                raise ValueError("This bus is not part of the pinmux")

            pin_candidates = [
                pin for pin in pin_candidates if pin not in self.configured
            ]

            if not pin_candidates:
                raise KeyErrorNotFound("No pin available for signal")

            if len(pin_candidates) == 1:
                self.set_function(pin_candidates[0], signal)
                continue

            if pins is None:
                pin = pin_candidates[0]
            else:
                try:
                    pin = find(pin_candidates, lambda pin: pin in pins)
                except KeyErrorAmbiguous as e:
                    pin = e.duplicates[0]

            self.set_function(pin, signal)



from abc import abstractmethod
from typing import cast

import pytest

from faebryk.core.node import Node, NodeAlreadyBound
from faebryk.core.trait import (
    Trait,
    TraitImpl,
    TraitImplementationConfusedWithTrait,
    TraitNotFound,
    TraitUnbound,
)


def test_trait_equality():
    class _trait1(Trait):
        pass

    class _trait1_1(_trait1):
        pass

    class _trait2(Trait):
        pass

    class impl1(_trait1.impl()):
        pass

    class impl1_1(impl1):
        pass

    class impl1_1_1(impl1_1):
        pass

    class impl1_2(impl1):
        pass

    class impl_1_1(_trait1_1.impl()):
        pass

    class impl2(_trait2.impl()):
        pass

    a = impl1()

    # Test instance and class equality
    assert impl1() != impl1()  # different instances
    assert a == a  # same instance
    assert impl1 == impl1  # same class
    assert impl1_1 != impl1  # class & parent/child class
    assert impl2 != impl1  # class & parallel class

    # Test trait implementation comparisons
    assert impl1().cmp(impl1())[0]  # inst & class
    assert impl1_1().cmp(impl1_1())[0]  # inst & class
    assert impl1_1().cmp(impl1())[0]  # inst & parent class
    assert impl1().cmp(impl1_1())[0]  # inst & child class
    assert not impl2().cmp(impl1())[0]  # inst & parallel class
    assert not impl2().cmp(impl1_1())[0]  # inst & parallel class
    assert impl1_1_1().cmp(impl1())[0]  # inst & double child class
    assert impl1().cmp(impl1_1_1())[0]  # inst & double parent class
    assert impl1_2().cmp(impl1_1())[0]  # inst & sister class
    assert impl1_2().cmp(impl1_1_1())[0]  # inst & nephew class

    # Test trait inheritance
    assert impl1().cmp(impl_1_1())[0]
    assert impl_1_1().cmp(impl1())[0]


def test_trait_basic_operations():
    obj = Node()

    class trait1(Trait):
        @abstractmethod
        def do(self) -> int:
            raise NotImplementedError

    class trait1impl(trait1.impl()):
        def do(self) -> int:
            return 1

    # Test failure on getting non-existent trait
    assert not obj.has_trait(trait1)
    with pytest.raises(TraitNotFound):
        obj.get_trait(trait1)

    # Test adding and getting trait
    trait1_inst = trait1impl()
    obj.add(trait1_inst)
    assert obj.has_trait(trait1)
    assert trait1_inst == obj.get_trait(trait1)
    assert trait1_inst.do() == obj.get_trait(trait1).do()

    # Test double add
    with pytest.raises(NodeAlreadyBound):
        obj.add(trait1_inst)

    # Test trait removal
    obj.del_trait(trait1)
    assert not obj.has_trait(trait1)


def test_trait_replacement():
    obj = Node()

    class trait1(Trait):
        @abstractmethod
        def do(self) -> int:
            raise NotImplementedError

    class trait1impl(trait1.impl()):
        def do(self) -> int:
            return 1

    class cfgtrait1(trait1impl):
        def __init__(self, cfg) -> None:
            super().__init__()
            self.cfg = cfg

        def do(self) -> int:
            return self.cfg

    trait1_inst = trait1impl()
    cfgtrait1_inst = cfgtrait1(5)

    # Test trait replacement
    obj.add(trait1_inst)
    obj.add(cfgtrait1_inst)
    assert cfgtrait1_inst == obj.get_trait(trait1)
    assert cfgtrait1_inst.do() == obj.get_trait(trait1).do()
    obj.add(trait1_inst)
    assert trait1_inst == obj.get_trait(trait1)


def test_trait_object_binding():
    obj = Node()

    class trait1(Trait):
        pass

    class trait1impl(trait1.impl()):
        pass

    trait1_inst = trait1impl()

    # Test trait object binding
    with pytest.raises(TraitUnbound):
        trait1_inst.obj
    obj.add(trait1_inst)
    _impl: TraitImpl = cast(TraitImpl, obj.get_trait(trait1))
    assert _impl.obj == obj
    obj.del_trait(trait1)
    with pytest.raises(TraitUnbound):
        trait1_inst.obj


def test_trait_inheritance_and_override():
    obj = Node()

    class trait1(Trait):
        pass

    class trait2(trait1):
        pass

    class impl1(trait1.impl()):
        pass

    class impl2(trait2.impl()):
        pass

    impl1_inst = impl1()
    impl2_inst = impl2()

    # Test specific override
    obj.add(impl2_inst)
    obj.add(impl1_inst)
    assert impl2_inst == obj.get_trait(trait1)

    # Test child delete
    obj.del_trait(trait1)
    assert not obj.has_trait(trait1)


def test_trait_impl_confusion():
    obj = Node()

    class trait1(Trait):
        pass

    class trait1impl(trait1.impl()):
        pass

    t1 = obj.add(trait1impl())
    with pytest.raises(TraitImplementationConfusedWithTrait):
        obj.get_trait(trait1impl)

    assert obj.get_trait(trait1) == t1


def test_trait_impl_exception():
    obj = Node()

    class trait1impl(Trait.decless()):
        pass

    # ensure it's a class attribute
    assert trait1impl.__trait__ is trait1impl().__trait__

    t1 = obj.add(trait1impl())
    assert obj.get_trait(trait1impl) is t1


def test_trait_decless_basic():
    class T1(Trait.decless()):
        def __init__(self, data: str) -> None:
            super().__init__()
            self.data = data

    n = Node()
    n.add(T1("test"))

    assert n.has_trait(T1)
    assert n.get_trait(T1).data == "test"


def test_trait_decless_inherit():
    class T1(Trait.decless()):
        def __init__(self, data: str) -> None:
            super().__init__()
            self.data = data

    class T2(T1):
        def __init__(self, data: str, more_data: str) -> None:
            super().__init__(data)
            self.more_data = more_data

    n = Node()
    n.add(T2("test", "more"))

    assert n.has_trait(T1)
    assert n.get_trait(T1).data == "test"
    assert n.has_trait(T2)
    assert n.get_trait(T2).more_data == "more"


def test_trait_decless_inherit_2():
    class T1(Trait):
        pass

    class T2(T1.decless()):
        def __init__(self, data: str) -> None:
            super().__init__()
            self.data = data

    n = Node()
    n.add(T2("test"))

    assert n.has_trait(T1)
    assert n.has_trait(T2)
    assert n.get_trait(T2).data == "test"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class HDMI(ModuleInterface):
    """
    HDMI interface
    """

    power: F.ElectricPower
    data = L.list_field(3, F.DifferentialPair)
    clock: F.DifferentialPair
    i2c: F.I2C
    cec: F.ElectricLogic
    hotplug: F.ElectricLogic

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    # @staticmethod
    # def define_max_frequency_capability(mode: SpeedMode):
    #     return F.Range(I2C.SpeedMode.low_speed, mode)

    def __preinit__(self) -> None:
        pass



from pathlib import Path

import pytest
from pygments import token as pygments_token

import atopile
import atopile.parse
import atopile.parse_utils
from faebryk.libs.util import repo_root as _repo_root


def _parser(src: str):
    input = atopile.parse.InputStream(src)
    input.name = "test"
    parser, _ = atopile.parse.make_parser(input)
    return parser


@pytest.mark.parametrize(
    "txt",
    [
        "a = 1",
        "assert 1 within 2",
        "b = 1kV +/- 10% + (2V - 1V)",
    ],
)
def test_reconstructor_simple_stmt(txt: str):
    """Ensure we can faithfully re-construct the source code from a parse tree"""
    assert atopile.parse_utils.reconstruct(_parser(txt).simple_stmt()) == txt


repo_root = _repo_root()
EXAMPLES_DIR = repo_root / "examples"


@pytest.mark.slow
@pytest.mark.parametrize(
    "example",
    [p for p in EXAMPLES_DIR.glob("*.ato") if p.is_file()],
    ids=lambda p: p.stem,
)
def test_example_reconstruction(example: Path):
    assert example.exists()
    file_ast = atopile.parse.parse_file(example)
    assert atopile.parse_utils.reconstruct(file_ast) == example.read_text(
        encoding="utf-8"
    )


def test_partial_reconstruction():
    code = "a=1;b=2"
    file_ast = atopile.parse.parse_text_as_file(code)
    assert (
        atopile.parse_utils.reconstruct(
            file_ast.stmt()[0].simple_stmts().simple_stmt()[1]
        )
        == "b=2"
    )


@pytest.fixture
def abcde():
    code = {
        "a": "a=1",
        "b": "b=2",
        "c": "c=3",
        "d": "d=4",
        "e": "e=5",
    }
    src = "a=1;b=2\nc=3;d=4\ne=5"
    file_ast = atopile.parse.parse_text_as_file(src)
    stmts = {
        "a": file_ast.stmt()[0].simple_stmts().simple_stmt()[0],
        "b": file_ast.stmt()[0].simple_stmts().simple_stmt()[1],
        "c": file_ast.stmt()[1].simple_stmts().simple_stmt()[0],
        "d": file_ast.stmt()[1].simple_stmts().simple_stmt()[1],
        "e": file_ast.stmt()[2].simple_stmts().simple_stmt()[0],
    }
    return file_ast, stmts, code


def test_reconstruct_expand(abcde):
    _, stmts, _ = abcde
    assert atopile.parse_utils.reconstruct(stmts["c"]) == "c=3"
    assert atopile.parse_utils.reconstruct(stmts["c"], expand_before=0) == "c=3"
    assert (
        atopile.parse_utils.reconstruct(stmts["c"], expand_before=1) == "a=1;b=2\nc=3"
    )
    assert atopile.parse_utils.reconstruct(stmts["c"], expand_after=0) == "c=3;d=4"
    assert atopile.parse_utils.reconstruct(stmts["c"], expand_after=1) == "c=3;d=4\ne=5"
    assert atopile.parse_utils.reconstruct(stmts["d"]) == "d=4"
    assert atopile.parse_utils.reconstruct(stmts["d"], expand_after=0) == "d=4"
    assert atopile.parse_utils.reconstruct(stmts["d"], expand_before=0) == "c=3;d=4"


def test_pygments_lexer():
    text = """module Test:
    a = 1
"""
    lexer = atopile.parse_utils.PygmentsLexer()
    tok_gen = lexer.get_tokens_unprocessed(text)
    assert list(filter(lambda t: t[1] != pygments_token.Whitespace, tok_gen)) == [
        (0, pygments_token.Keyword, "module"),
        (7, pygments_token.Name, "Test"),
        (11, pygments_token.Token, ":"),
        (17, pygments_token.Name, "a"),
        (19, pygments_token.Operator, "="),
        (21, pygments_token.Number, "1"),
    ]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging

from faebryk.core.solver.solver import Solver
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


class has_solver(Trait.decless()):
    def __init__(self, solver: Solver):
        self._solver = solver
        super().__init__()

    @property
    def solver(self) -> Solver:
        return self._solver



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
WARNING: Only use this module in the library and apps!
"""

import logging
from typing import Any

from faebryk.core.module import Module  # noqa: F401
from faebryk.core.moduleinterface import ModuleInterface  # noqa: F401
from faebryk.core.node import (  # noqa: F401
    InitVar,
    Node,
    d_field,
    f_field,
    list_f_field,
    list_field,
    rt_field,
)
from faebryk.core.parameter import R, p_field  # noqa: F401
from faebryk.core.reference import reference  # noqa: F401
from faebryk.core.trait import Trait  # noqa: F401
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
    Quantity_Set_Discrete,
    Quantity_Set_Empty,
    Quantity_Singleton,
    QuantityLike,
    QuantityLikeR,
)
from faebryk.libs.sets.sets import BoolSet, EnumSet  # noqa: F401
from faebryk.libs.units import Unit


class AbstractclassError(Exception): ...


logger = logging.getLogger(__name__)


Predicates = R.Predicates
Domains = R.Domains
Expressions = R.Expressions

# TODO: Pint is really bad with types
# Thus we need to allow everything to avoid type errors all over the place
# This is really annoying and should be fixed in Pint
# 5 * P.V is Quantity | Unknown | PlainQuantity[Any]

type RelaxedQuantity = QuantityLike | Any


class Range(Quantity_Interval):
    def __init__(
        self,
        min: RelaxedQuantity | None = None,
        max: RelaxedQuantity | None = None,
        units: Unit | None = None,
    ):
        if min is not None:
            assert isinstance(min, QuantityLikeR)
        if max is not None:
            assert isinstance(max, QuantityLikeR)
        super().__init__(min, max, units)


class RangeWithGaps(Quantity_Interval_Disjoint):
    def __init__(
        self,
        *intervals: "Quantity_Interval | Quantity_Interval_Disjoint | tuple[RelaxedQuantity, RelaxedQuantity]",  # noqa: E501
        units: Unit | None = None,
    ):
        super().__init__(*intervals, units=units)


class Single(Quantity_Singleton):
    def __init__(self, value: RelaxedQuantity):
        assert isinstance(value, QuantityLikeR)
        super().__init__(value)


class DiscreteSet(Quantity_Set_Discrete):
    def __init__(self, *values: RelaxedQuantity, units: Unit | None = None):
        assert all(isinstance(v, QuantityLikeR) for v in values)
        super().__init__(*values, units=units)


def EmptySet(units: Unit | None = None):
    return Quantity_Set_Empty(units)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class Regulator(Module):
    power_in: F.ElectricPower
    power_out: F.ElectricPower

    def __preinit__(self):
        self.power_out.add(F.Power.is_power_source.impl()())
        self.power_in.add(F.Power.is_power_sink.impl()())

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.power_in, self.power_out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file is autogenerated by tools/library/gen_F.py
This is the __init__.py file of the library
All modules are in ./<module>.py with name class <module>
Export all <module> classes here
Do it programmatically instead of specializing each manually
This way we can add new modules without changing this file
"""

# Disable ruff warning for whole block
# flake8: noqa: F401
# flake8: noqa: I001
# flake8: noqa: E501

from faebryk.library.Electrical import Electrical
from faebryk.library.has_designator_prefix import has_designator_prefix
from faebryk.library.has_package_requirements import has_package_requirements
from faebryk.library.can_specialize_defined import can_specialize_defined
from faebryk.library.Power import Power
from faebryk.library.can_be_surge_protected import can_be_surge_protected
from faebryk.library.has_net_name import has_net_name
from faebryk.library.is_bus_parameter import is_bus_parameter
from faebryk.library.Signal import Signal
from faebryk.library.has_single_electric_reference import has_single_electric_reference
from faebryk.library.is_optional_defined import is_optional_defined
from faebryk.library.has_footprint import has_footprint
from faebryk.library.Mechanical import Mechanical
from faebryk.library.has_overriden_name import has_overriden_name
from faebryk.library.has_reference import has_reference
from faebryk.library.has_linked_pad import has_linked_pad
from faebryk.library.has_pcb_position import has_pcb_position
from faebryk.library.has_part_picked import has_part_picked
from faebryk.library.is_auto_generated import is_auto_generated
from faebryk.library.is_app_root import is_app_root
from faebryk.library.has_layout_transform import has_layout_transform
from faebryk.library.has_solver import has_solver
from faebryk.library.can_bridge import can_bridge
from faebryk.library.can_specialize import can_specialize
from faebryk.library.has_datasheet import has_datasheet
from faebryk.library.has_designator import has_designator
from faebryk.library.has_esphome_config import has_esphome_config
from faebryk.library.is_lazy import is_lazy
from faebryk.library.has_descriptive_properties import has_descriptive_properties
from faebryk.library.has_simple_value_representation import has_simple_value_representation
from faebryk.library.has_capacitance import has_capacitance
from faebryk.library.has_construction_dependency import has_construction_dependency
from faebryk.library.has_kicad_ref import has_kicad_ref
from faebryk.library.has_pcb_layout import has_pcb_layout
from faebryk.library.has_pcb_routing_strategy import has_pcb_routing_strategy
from faebryk.library.is_optional import is_optional
from faebryk.library.has_reference_layout import has_reference_layout
from faebryk.library.has_resistance import has_resistance
from faebryk.library.has_single_connection import has_single_connection
from faebryk.library.is_esphome_bus import is_esphome_bus
from faebryk.library.is_not_pickable import is_not_pickable
from faebryk.library.is_pickable import is_pickable
from faebryk.library.is_representable_by_single_value import is_representable_by_single_value
from faebryk.library.XtalIF import XtalIF
from faebryk.library.has_pin_association_heuristic import has_pin_association_heuristic
from faebryk.library.Filter import Filter
from faebryk.library.Logic import Logic
from faebryk.library.has_single_electric_reference_defined import has_single_electric_reference_defined
from faebryk.library.Footprint import Footprint
from faebryk.library.has_overriden_name_defined import has_overriden_name_defined
from faebryk.library.Symbol import Symbol
from faebryk.library.has_linked_pad_defined import has_linked_pad_defined
from faebryk.library.has_pcb_position_defined import has_pcb_position_defined
from faebryk.library.has_pcb_position_defined_relative import has_pcb_position_defined_relative
from faebryk.library.has_pcb_position_defined_relative_to_parent import has_pcb_position_defined_relative_to_parent
from faebryk.library.has_part_removed import has_part_removed
from faebryk.library.implements_design_check import implements_design_check
from faebryk.library.can_bridge_by_name import can_bridge_by_name
from faebryk.library.can_bridge_defined import can_bridge_defined
from faebryk.library.has_datasheet_defined import has_datasheet_defined
from faebryk.library.has_esphome_config_defined import has_esphome_config_defined
from faebryk.library.has_descriptive_properties_defined import has_descriptive_properties_defined
from faebryk.library.has_simple_value_representation_based_on_params import has_simple_value_representation_based_on_params
from faebryk.library.has_simple_value_representation_based_on_params_chain import has_simple_value_representation_based_on_params_chain
from faebryk.library.has_simple_value_representation_defined import has_simple_value_representation_defined
from faebryk.library.has_pcb_layout_defined import has_pcb_layout_defined
from faebryk.library.has_single_connection_impl import has_single_connection_impl
from faebryk.library.is_esphome_bus_defined import is_esphome_bus_defined
from faebryk.library.is_pickable_by_part_number import is_pickable_by_part_number
from faebryk.library.is_pickable_by_supplier_id import is_pickable_by_supplier_id
from faebryk.library.is_pickable_by_type import is_pickable_by_type
from faebryk.library.is_representable_by_single_value_defined import is_representable_by_single_value_defined
from faebryk.library.has_pin_association_heuristic_lookup_table import has_pin_association_heuristic_lookup_table
from faebryk.library.LogicOps import LogicOps
from faebryk.library.can_attach_to_footprint import can_attach_to_footprint
from faebryk.library.can_attach_via_pinmap import can_attach_via_pinmap
from faebryk.library.has_footprint_impl import has_footprint_impl
from faebryk.library.has_kicad_footprint import has_kicad_footprint
from faebryk.library.has_symbol_layout import has_symbol_layout
from faebryk.library.Pad import Pad
from faebryk.library.PCB import PCB
from faebryk.library.requires_external_usage import requires_external_usage
from faebryk.library.GDT import GDT
from faebryk.library.BJT import BJT
from faebryk.library.Diode import Diode
from faebryk.library.MOSFET import MOSFET
from faebryk.library.LogicGate import LogicGate
from faebryk.library.has_footprint_defined import has_footprint_defined
from faebryk.library.has_symbol_layout_defined import has_symbol_layout_defined
from faebryk.library.Net import Net
from faebryk.library.can_attach_via_pinmap_pinlist import can_attach_via_pinmap_pinlist
from faebryk.library.has_equal_pins import has_equal_pins
from faebryk.library.has_kicad_manual_footprint import has_kicad_manual_footprint
from faebryk.library.has_pcb_routing_strategy_greedy_direct_line import has_pcb_routing_strategy_greedy_direct_line
from faebryk.library.TVS import TVS
from faebryk.library.NFET import NFET
from faebryk.library.PFET import PFET
from faebryk.library.LogicGates import LogicGates
from faebryk.library.can_attach_to_footprint_symmetrically import can_attach_to_footprint_symmetrically
from faebryk.library.can_attach_to_footprint_via_pinmap import can_attach_to_footprint_via_pinmap
from faebryk.library.has_pcb_routing_strategy_manual import has_pcb_routing_strategy_manual
from faebryk.library.has_pcb_routing_strategy_via_to_layer import has_pcb_routing_strategy_via_to_layer
from faebryk.library.can_attach_via_pinmap_equal import can_attach_via_pinmap_equal
from faebryk.library.has_equal_pins_in_ifs import has_equal_pins_in_ifs
from faebryk.library.has_kicad_footprint_equal_ifs import has_kicad_footprint_equal_ifs
from faebryk.library.KicadFootprint import KicadFootprint
from faebryk.library.SurgeProtection import SurgeProtection
from faebryk.library.Capacitor import Capacitor
from faebryk.library.Crystal import Crystal
from faebryk.library.Fuse import Fuse
from faebryk.library.Inductor import Inductor
from faebryk.library.Resistor import Resistor
from faebryk.library.Switch import Switch
from faebryk.library.DIP import DIP
from faebryk.library.QFN import QFN
from faebryk.library.SMDTwoPin import SMDTwoPin
from faebryk.library.SOIC import SOIC
from faebryk.library.has_kicad_footprint_equal_ifs_defined import has_kicad_footprint_equal_ifs_defined
from faebryk.library.Mounting_Hole import Mounting_Hole
from faebryk.library.TestPoint import TestPoint
from faebryk.library.has_explicit_part import has_explicit_part
from faebryk.library.is_atomic_part import is_atomic_part
from faebryk.library.is_surge_protected import is_surge_protected
from faebryk.library.CapacitorElectrolytic import CapacitorElectrolytic
from faebryk.library.MultiCapacitor import MultiCapacitor
from faebryk.library.is_decoupled import is_decoupled
from faebryk.library.Common_Mode_Filter import Common_Mode_Filter
from faebryk.library.Crystal_Oscillator import Crystal_Oscillator
from faebryk.library.Potentiometer import Potentiometer
from faebryk.library.ResistorArray import ResistorArray
from faebryk.library.is_surge_protected_defined import is_surge_protected_defined
from faebryk.library.can_be_decoupled import can_be_decoupled
from faebryk.library.can_be_surge_protected_defined import can_be_surge_protected_defined
from faebryk.library.ElectricPower import ElectricPower
from faebryk.library.Battery import Battery
from faebryk.library.Comparator import Comparator
from faebryk.library.ElectricSignal import ElectricSignal
from faebryk.library.Fan import Fan
from faebryk.library.LED import LED
from faebryk.library.NetTie import NetTie
from faebryk.library.OpAmp import OpAmp
from faebryk.library.Regulator import Regulator
from faebryk.library.Relay import Relay
from faebryk.library.can_be_decoupled_rails import can_be_decoupled_rails
from faebryk.library.DifferentialPair import DifferentialPair
from faebryk.library.ElectricLogic import ElectricLogic
from faebryk.library.FilterElectricalLC import FilterElectricalLC
from faebryk.library.FilterElectricalRC import FilterElectricalRC
from faebryk.library.PowerMux import PowerMux
from faebryk.library.ResistorVoltageDivider import ResistorVoltageDivider
from faebryk.library.has_single_electric_reference_shared import has_single_electric_reference_shared
from faebryk.library.requires_pulls import requires_pulls
from faebryk.library.PoweredLED import PoweredLED
from faebryk.library.RS485HalfDuplex import RS485HalfDuplex
from faebryk.library.Addressor import Addressor
from faebryk.library.ElectricLogicGate import ElectricLogicGate
from faebryk.library.EnablePin import EnablePin
from faebryk.library.Ethernet import Ethernet
from faebryk.library.GenericBusProtection import GenericBusProtection
from faebryk.library.I2S import I2S
from faebryk.library.JTAG import JTAG
from faebryk.library.MultiSPI import MultiSPI
from faebryk.library.PDM import PDM
from faebryk.library.Pinmux import Pinmux
from faebryk.library.RS232 import RS232
from faebryk.library.SPI import SPI
from faebryk.library.SWD import SWD
from faebryk.library.Sercom import Sercom
from faebryk.library.UART_Base import UART_Base
from faebryk.library.USB2_0_IF import USB2_0_IF
from faebryk.library.can_switch_power import can_switch_power
from faebryk.library.Resistor_Voltage_Divider import Resistor_Voltage_Divider
from faebryk.library.I2C import I2C
from faebryk.library.ElectricLogicGates import ElectricLogicGates
from faebryk.library.LDO import LDO
from faebryk.library.SPIFlash import SPIFlash
from faebryk.library.UART import UART
from faebryk.library.USB2_0 import USB2_0
from faebryk.library.USB3_IF import USB3_IF
from faebryk.library.can_switch_power_defined import can_switch_power_defined
from faebryk.library.EEPROM import EEPROM
from faebryk.library.HDMI import HDMI
from faebryk.library.USB3 import USB3
from faebryk.library.PowerSwitch import PowerSwitch
from faebryk.library.USB_C import USB_C
from faebryk.library.PowerSwitchMOSFET import PowerSwitchMOSFET
from faebryk.library.PowerSwitchStatic import PowerSwitchStatic
from faebryk.library.LEDIndicator import LEDIndicator



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import re
import subprocess
from pathlib import Path


class GithubException(Exception):
    pass


class GithubCLINotFound(GithubException):
    pass


class GithubUserNotLoggedIn(GithubException):
    pass


class GithubRepoAlreadyExists(GithubException):
    pass


class GithubRepoNotFound(GithubException):
    pass


class GithubCLI:
    def __init__(self):
        try:
            subprocess.check_output(["gh", "--version"])
        except Exception:
            raise GithubCLINotFound()

        # check if logged in
        self.get_usernames()

    @property
    def use_ssh(self) -> bool:
        # BUG
        # subprocess.check_output(
        #    ["gh", "config", "get", "git_protocol"], stderr=subprocess.DEVNULL
        # )
        configs = (
            subprocess.check_output(["gh", "config", "list"], stderr=subprocess.DEVNULL)
            .decode(encoding="utf-8")
            .splitlines()
        )
        return "git_protocol=ssh" in configs

    def get_usernames(self) -> list[str]:
        # github.com
        #  ✓ Logged in to github.com account <USERNAME> (~/.config/gh/hosts.yml)
        #  - Active account: true
        #  - Git operations protocol: ssh
        #  - Token: gho_************************************
        #  - Token scopes: 'admin:public_key', 'gist', 'read:org', 'repo'
        #
        #  ✓ Logged in to github.com account <USERNAME> (~/.config/gh/hosts.yml)
        #  - Active account: false
        #  - Git operations protocol: ssh
        #  - Token: gho_************************************
        #  - Token scopes: 'admin:public_key', 'gist', 'read:org', 'repo'

        try:
            auth = subprocess.check_output(
                ["gh", "auth", "status"], stderr=subprocess.DEVNULL
            ).decode(encoding="utf-8")
        except subprocess.CalledProcessError:
            raise GithubUserNotLoggedIn()
        logged_in = [
            line.strip() for line in auth.split("\n") if "Logged in to" in line
        ]
        users = [
            match.group(1)
            for line in logged_in
            if (match := re.match(r"^.*account\s+(\w+)\s+.*$", line))
        ]
        if not users:
            raise GithubUserNotLoggedIn()
        return users

    def get_orgs(self) -> list[str]:
        # Showing 2 of 2 organizations

        # faebryk
        # atopile

        orgs_res = subprocess.check_output(["gh", "org", "list"]).decode(
            encoding="utf-8"
        )
        orgs = [
            stripped
            for line in orgs_res.split("\n")
            if (stripped := line.strip()) and not stripped.startswith("Showing")
        ]

        user_orgs = self.get_usernames()
        return orgs + user_orgs

    def create_repo(
        self,
        repo_id: str,
        visibility: str = "public",
        add_remote: bool = False,
        path: Path | None = None,
    ) -> str:
        """
        Creates a GitHub repository.
        repo_id: The repository identifier, e.g., "owner/repo_name".
        visibility: "public" or "private".
        add_remote: If True, adds the new repo as a remote named 'origin'
        to the local git repo in CWD and pushes.
        Returns the URL of the created repository.
        """
        cmd = ["gh", "repo", "create", repo_id, f"--{visibility}"]
        if add_remote:
            cmd.extend(["--source=.", "--remote=origin", "--push"])

        try:
            # gh repo create prints the URL of the new repo to stdout
            result = subprocess.run(
                cmd, capture_output=True, text=True, check=True, cwd=path
            )
            # The URL is usually the last line of the output if successful
            url = result.stdout.strip().split("\\n")[-1]
            if not url.startswith("https://"):  # simple check if it's a URL
                # if --source=. is used, the output is just the URL
                # if not, it's like
                # "✓ Created repository <user>/<repo> on GitHub\n<URL>"
                match = re.search(r"(https://github.com/[\w-]+/[\w.-]+)", result.stdout)
                if match:
                    url = match.group(1)
                else:
                    raise GithubException(
                        f"Could not parse URL from gh repo create output:"
                        f" {result.stdout}"
                    )
            return url
        except subprocess.CalledProcessError as e:
            if "already exists" in e.stderr:
                raise GithubRepoAlreadyExists(
                    f"Repository {repo_id} already exists."
                ) from e
            raise GithubException(
                f"Failed to create repository {repo_id}: {e.stderr}"
            ) from e

    def get_repo_url(self, repo_id: str) -> str:
        """
        Gets the URL of an existing GitHub repository.
        repo_id: The repository identifier, e.g., "owner/repo_name".
        Returns the URL of the repository.
        """
        if self.use_ssh:
            cmd = ["gh", "repo", "view", repo_id, "--json", "sshUrl", "-q", ".sshUrl"]
        else:
            cmd = ["gh", "repo", "view", repo_id, "--json", "url", "-q", ".url"]
        try:
            result = subprocess.run(
                cmd, capture_output=True, text=True, check=True, cwd=None
            )
            return result.stdout.strip()
        except subprocess.CalledProcessError as e:
            if (
                "Could not find repository" in e.stderr
                or "Could not resolve" in e.stderr
            ):
                raise GithubRepoNotFound(f"Repository {repo_id} not found.") from e
            raise GithubException(
                f"Failed to get URL for repository {repo_id}: {e.stderr}"
            ) from e



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import tempfile
from pathlib import Path

from kicadcliwrapper.generated.kicad_cli import kicad_cli as k

from faebryk.libs.kicad.fileformats_latest import C_kicad_drc_report_file


def run_drc(pcb: Path):
    drc_report = None
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        out = tmpdir / "drc.json"
        k(
            k.pcb(
                k.pcb.drc(
                    INPUT_FILE=str(pcb),
                    format="json",
                    output=str(out),
                    severity_all=True,
                    schematic_parity=False,
                )
            )
        ).exec(check=True)

        drc_report = C_kicad_drc_report_file.loads(out)

    return drc_report



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class can_attach_to_footprint_via_pinmap(F.can_attach_to_footprint.impl()):
    def __init__(
        self, pinmap: dict[str, F.Electrical | None] | dict[str, F.Electrical]
    ) -> None:
        super().__init__()
        self.pinmap: dict[str, F.Electrical | None] = pinmap  # type: ignore

    def attach(self, footprint: F.Footprint):
        self.obj.add(F.has_footprint_defined(footprint))
        footprint.get_trait(F.can_attach_via_pinmap).attach(self.pinmap)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
import math
from itertools import combinations
from typing import cast

from faebryk.core.parameter import (
    Add,
    CanonicalExpression,
    ConstrainableExpression,
    Domain,
    Expression,
    GreaterOrEqual,
    HasSideEffects,
    Is,
    IsSubset,
    Multiply,
    Parameter,
    ParameterOperatable,
    Power,
)
from faebryk.core.solver.algorithm import algorithm
from faebryk.core.solver.mutator import Mutator
from faebryk.core.solver.utils import (
    Contradiction,
    ContradictionByLiteral,
    SolverLiteral,
    make_lit,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
)
from faebryk.libs.sets.sets import BoolSet, P_Set
from faebryk.libs.util import (
    EquivalenceClasses,
    cast_assert,
    groupby,
)

logger = logging.getLogger(__name__)


# TODO: mark terminal=False where applicable


@algorithm("Check literal contradiction", terminal=False)
def check_literal_contradiction(mutator: Mutator):
    """
    Check if a literal is contradictory
    """

    mutator.get_literal_mappings(new_only=False, allow_subset=True)


@algorithm("Convert inequality with literal to subset", terminal=False)
def convert_inequality_with_literal_to_subset(mutator: Mutator):
    # TODO if not! A <= x it can be replaced by A intersect [-inf, a] is {}
    """
    This only works for inequalities we know cannot evaluate to {True, False}
    A >=! 5 -> A ss! [5, inf)
    5 >=! A -> A ss! (-inf, 5]

    A >=! [1, 10] -> A ss! [10, inf)
    [1, 10] >=! A -> A ss! (-inf, 1]
    """

    ge_exprs = {
        e
        for e in mutator.nodes_of_type(GreaterOrEqual, sort_by_depth=True)
        # Look for expressions with only one non-literal operand
        if e.constrained
        and len(list(op for op in e.operands if isinstance(op, ParameterOperatable)))
        == 1
    }

    for ge in ge_exprs:
        is_left = ge.operands[0] is next(iter(ge.operatable_operands))

        if is_left:
            param = ge.operands[0]
            lit = Quantity_Interval_Disjoint.from_value(ge.operands[1])
            boundary = lit.max_elem
            if math.isinf(boundary):
                if ge.constrained:
                    raise Contradiction(
                        "GreaterEqual inf not possible",
                        involved=[param],
                        mutator=mutator,
                    )
                mutator.utils.alias_is_literal_and_check_predicate_eval(ge, False)
                continue
            interval = Quantity_Interval_Disjoint(Quantity_Interval(min=boundary))
        else:
            param = ge.operands[1]
            lit = Quantity_Interval_Disjoint.from_value(ge.operands[0])
            boundary = lit.min_elem
            if math.isinf(boundary):
                if ge.constrained:
                    raise Contradiction(
                        "LessEqual -inf not possible",
                        involved=[param],
                        mutator=mutator,
                    )
                mutator.utils.alias_is_literal_and_check_predicate_eval(ge, False)
                continue
            interval = Quantity_Interval_Disjoint(Quantity_Interval(max=boundary))

        mutator.mutate_expression(
            ge,
            operands=[param, interval],
            expression_factory=IsSubset,
        )


@algorithm("Remove unconstrained", terminal=True)
def remove_unconstrained(mutator: Mutator):
    """
    Remove all expressions that are not involved in any constrained predicates
    or expressions with side effects
    Note: Not possible for Parameters, want to keep those around for REPR
    """
    objs = mutator.nodes_of_type(Expression)
    for obj in objs:
        if obj.constrained:
            continue
        if isinstance(obj, HasSideEffects):
            continue
        if any(
            e.constrained or isinstance(e, HasSideEffects)
            for e in mutator.utils.get_expressions_involved_in(obj)
        ):
            continue
        mutator.remove(obj)


@algorithm("Remove congruent expressions", terminal=False)
def remove_congruent_expressions(mutator: Mutator):
    """
    Remove expressions that are congruent to other expressions

    ```
    X1 := A + B, X2 := A + B -> [{X1, X2}, {A}, {B}]
    ```
    """
    # X1 = A + B, X2 = A + B -> X1 is X2
    # No (Invalid): X1 = A + [0, 10], X2 = A + [0, 10]
    # No (Automatic): X1 = A + C, X2 = A + B, C ~ B -> X1 ~ X2

    all_exprs = mutator.nodes_of_type(Expression, sort_by_depth=True)
    # optimization: can't be congruent if they have uncorrelated literals
    all_exprs = [e for e in all_exprs if not e.get_uncorrelatable_literals()]
    # TODO is this fully correct?
    # optimization: Is, IsSubset already handled
    all_exprs = [
        e
        for e in all_exprs
        if not (
            isinstance(e, (Is, IsSubset)) and e.constrained and e.get_operand_literals()
        )
    ]
    exprs_by_type = groupby(
        all_exprs,
        lambda e: (
            type(e),
            len(e.operands),
            None if not isinstance(e, ConstrainableExpression) else e.constrained,
        ),
    )
    full_eq = EquivalenceClasses[Expression](all_exprs)

    for exprs in exprs_by_type.values():
        if len(exprs) <= 1:
            continue
        # TODO use hash to speed up comparisons
        for e1, e2 in combinations(exprs, 2):
            # no need for recursive, since subexpr already merged if congruent
            if not full_eq.is_eq(e1, e2) and e1.is_congruent_to(e2, recursive=False):
                full_eq.add_eq(e1, e2)

    repres = {}
    for expr in all_exprs:
        eq_class = full_eq.classes[expr]
        if len(eq_class) <= 1:
            continue

        eq_id = id(eq_class)
        if eq_id not in repres:
            representative = mutator.mutate_expression(expr)
            repres[eq_id] = representative

            # propagate constrained & terminate
            if isinstance(representative, ConstrainableExpression):
                eq_class = cast(set[ConstrainableExpression], eq_class)
                representative.constrained = any(e.constrained for e in eq_class)
                if any(mutator.is_predicate_terminated(e) for e in eq_class):
                    mutator.predicate_terminate(representative)

        mutator._mutate(expr, repres[eq_id])


@algorithm("Alias classes", terminal=False)
def resolve_alias_classes(mutator: Mutator):
    """
    Resolve alias classes
    Ignores literals
    ```
    A alias B
    B alias C
    C alias D + E
    D + E < 5
    D + E is 3
    D + E + F < 10
    -> A,B,C => R, R alias D + E, R < 5, R is 3, R + F < 10

    # Notes:
    D is 2, E is 5 -> D + E is 7, D + E is R, R is 3
    (D + E) - (D + E) -> R - R
    (D + E) + E -> R + E

    Careful: Aliases to literals will not be resolved due to loss
    of correlation information
    ```
    """

    # A is B, B is C, D is E, F, G is (A+B)
    # -> [{A, B, C}, {D, E}, {F}, {G, (A+B)}]
    param_ops = mutator.nodes_of_type(ParameterOperatable)
    full_eq = EquivalenceClasses[ParameterOperatable](param_ops)
    is_exprs = mutator.utils.get_all_aliases()
    for is_expr in is_exprs:
        ops = {
            # Literal expressions are basically literals
            o
            for o in is_expr.get_operand_operatables()
            if not mutator.utils.is_literal_expression(o)
        }
        # eq between non-literal operands
        full_eq.add_eq(*ops)
    p_eq_classes = full_eq.get(only_multi=True)

    # Make new param repre for alias classes
    for eq_class in p_eq_classes:
        eq_class_params = [p for p in eq_class if isinstance(p, Parameter)]
        eq_class_exprs = {p for p in eq_class if isinstance(p, Expression)}

        if not eq_class_params:
            continue

        if len(eq_class_params) == 1:
            # check if all in eq_class already aliased
            # Then no need to to create new representative
            _repr = eq_class_params[0]
            iss = _repr.get_operations(Is)
            iss_exprs = {
                o
                for e in iss
                for o in e.operatable_operands
                if isinstance(o, Expression)
            }
            if eq_class_exprs.issubset(iss_exprs):
                # check if all predicates are already propagated
                class_expressions = {
                    e
                    for operand in eq_class_exprs
                    for e in operand.get_operations()
                    # skip POps Is, because they create the alias classes
                    # or literal aliases (done by distribute algo)
                    if not (isinstance(e, Is) and e.constrained)
                    # skip literal subsets (done by distribute algo)
                    and not (
                        isinstance(e, IsSubset)
                        and e.get_operand_literals()
                        and e.constrained
                    )
                }
                if not class_expressions:
                    continue
            # else
            mutator.get_copy(_repr)
            continue

        # Merge param alias classes
        representative = mutator.utils.merge_parameters(eq_class_params)

        for p in eq_class_params:
            mutator._mutate(p, representative)

    for eq_class in p_eq_classes:
        eq_class_params = [p for p in eq_class if isinstance(p, Parameter)]
        eq_class_exprs = [p for p in eq_class if isinstance(p, Expression)]

        # single domain
        domain = Domain.get_shared_domain(*(op.domain for op in eq_class))

        if eq_class_params:
            # See len(alias_class_params) == 1 case above
            if not mutator.has_been_mutated(eq_class_params[0]):
                continue
            representative = mutator.get_mutated(eq_class_params[0])
        else:
            # If not params or lits in class, create a new param as representative
            # for expressions
            representative = mutator.register_created_parameter(
                Parameter(domain=domain), from_ops=list(eq_class)
            )

        for e in eq_class_exprs:
            mutator.soft_replace(e, representative)
            if mutator.utils.are_aliased(e, *eq_class_params):
                continue
            mutator.utils.alias_to(e, representative, from_ops=list(eq_class))


@algorithm("Merge intersecting subsets", terminal=False)
def merge_intersect_subsets(mutator: Mutator):
    """
    A subset L1
    A subset L2
    -> A subset (L1 & L2)

    x = A subset L1
    y = A subset L2
    Z = x and y -> Z = x & y
    -> A subset (L1 & L2)
    """

    # TODO use Intersection/Union expr
    # A ss B, A ss C -> A ss (B & C)
    # and then use literal folding
    # should also work for others:
    #   A < B, A < C -> A < (B | C)
    # Got to consider when to Intersect/Union is more useful than the associative

    # this merge is already done implicitly by try_extract_literal
    # but it's still needed to create the explicit subset op

    params = mutator.nodes_of_type(ParameterOperatable, sort_by_depth=True)

    for param in params:
        ss_lits = {
            k: vs
            for k, vs in mutator.utils.get_supersets(param).items()
            if mutator.utils.is_literal(k)
        }
        if len(ss_lits) <= 1:
            continue

        intersected = P_Set.intersect_all(*ss_lits.keys())

        # short-cut, would be detected by subset_to
        if intersected.is_empty():
            raise ContradictionByLiteral(
                "Intersection of literals is empty",
                involved=[param],
                literals=list(ss_lits.keys()),
                mutator=mutator,
            )

        old_ss = [old_ss for old_sss in ss_lits.values() for old_ss in old_sss]

        # already exists
        if intersected in ss_lits:
            target = ss_lits[intersected][0]
        else:
            target = mutator.utils.subset_to(param, intersected, from_ops=old_ss)
            assert isinstance(target, (IsSubset, Is))

        # Merge
        for old_ss in old_ss:
            mutator._mutate(old_ss, mutator.get_copy(target))


@algorithm("Empty set", terminal=False)
def empty_set(mutator: Mutator):
    """
    A is {} -> False
    A ss {} -> False
    """

    # A is {} -> False
    for e in mutator.nodes_of_type(Is):
        lits = cast(dict[int, SolverLiteral], e.get_operand_literals())
        if not lits:
            continue
        if any(lit.is_empty() for lit in lits.values()):
            mutator.utils.alias_is_literal_and_check_predicate_eval(e, False)

    # A ss {} -> False
    # Converted by literal_folding


@algorithm("Transitive subset", terminal=False)
def transitive_subset(mutator: Mutator):
    """
    ```
    A ss! B, B ss! C -> new A ss! C
    A ss! B, B is! X -> new A ss! X
    ```
    """
    # for all A ss! B | B not lit
    for ss_op in mutator.utils.get_all_subsets():
        A, B = ss_op.operands
        if not isinstance(B, ParameterOperatable):
            continue

        # all B ss! C | C not A
        for C, es in mutator.utils.get_supersets(B).items():
            if C is A:
                continue
            # create A ss! C/X
            mutator.utils.subset_to(A, C, from_ops=[ss_op, *es])

        # all B is! X, X lit
        # for non-lits done by eq classes
        X = mutator.utils.try_extract_literal(B)
        if X is not None:
            mutator.utils.subset_to(A, X, from_ops=[ss_op, B])


@algorithm("Predicate flat terminate", terminal=False)
def predicate_flat_terminate(mutator: Mutator):
    """
    ```
    P!(A, Lit) -> P!!(A, Lit)
    P!(Lit1, Lit2) -> P!!(Lit1, Lit2)
    ```

    Terminates all (dis)proven predicates that contain no expressions.
    """
    predicates = mutator.nodes_of_type(ConstrainableExpression)
    for p in predicates:
        if not p.constrained:
            continue

        if any(isinstance(po, Expression) for po in p.operatable_operands):
            continue

        # only (dis)proven
        if mutator.utils.try_extract_literal(p) is None:
            continue

        mutator.predicate_terminate(p)


@algorithm("Predicate is!! True", terminal=False)
def predicate_terminated_is_true(mutator: Mutator):
    """
    P!! is! True -> P!! is!! True
    """

    for p in mutator.nodes_of_type(Is):
        if not p.constrained:
            continue
        if mutator.is_predicate_terminated(p):
            continue
        if make_lit(True) not in p.operands:
            continue
        if not p.operatable_operands:
            continue
        op = next(iter(p.operatable_operands))
        if not op.constrained or not mutator.is_predicate_terminated(op):
            continue

        mutator.predicate_terminate(p)


@algorithm("Convert aliased singletons into literals", terminal=False)
def convert_operable_aliased_to_single_into_literal(mutator: Mutator):
    """
    A is ([5]), A + B -> ([5]) + B
    A is [True], A ^ B -> [True] ^ B
    """

    exprs = mutator.nodes_of_type(Expression, sort_by_depth=True)
    for e in exprs:
        if mutator.utils.is_pure_literal_expression(e):
            continue
        # handled in _todo
        if mutator.utils.is_alias_is_literal(e) or mutator.utils.is_subset_literal(e):
            continue
        # not handling here
        if (
            isinstance(e, (Is, IsSubset))
            and e.constrained
            and any(mutator.utils.is_constrained(op) for op in e.operatable_operands)
        ):
            continue

        ops = []
        found_literal = False
        for op in e.operands:
            lit = mutator.utils.is_replacable_by_literal(op)
            # preserve non-replaceable operands
            # A + B + C | A is ([5]) -> B, C
            if lit is None:
                ops.append(op)
                continue
            ops.append(lit)
            found_literal = True

        if not found_literal:
            continue

        mutator.mutate_expression(e, operands=ops)


@algorithm("Isolate lone parameters", terminal=False)
def isolate_lone_params(mutator: Mutator):
    """
    If an expression is aliased to a literal, and only one parameter in the expression
    is not aliased to a literal, isolate the lone parameter on one side of the
    expression.

    Inversion operations must be constructed using canonical operators only.

    A + B is Lit1, B is Lit2, A further uncorrelated B -> A is Lit1 + (Lit2 * -1)
    """

    def _isolate_param(
        param: ParameterOperatable,
        op_with_param: ParameterOperatable,
        op_without_param: ParameterOperatable,
        from_expr: Expression,
    ) -> tuple[ParameterOperatable.All, ParameterOperatable.All]:
        if not isinstance(op_with_param, Expression):
            return op_with_param, op_without_param

        def op_or_create_expr(
            operation: type[CanonicalExpression], *operands: ParameterOperatable.All
        ) -> ParameterOperatable.All:
            if len(operands) == 1:
                return operands[0]

            return mutator.create_expression(operation, *operands, from_ops=[from_expr])

        retained_ops = [
            op
            for op in op_with_param.operands
            if param in mutator.utils.find_unique_params(op)
        ]

        moved_ops = [
            op
            for op in op_with_param.operands
            if param not in mutator.utils.find_unique_params(op)
        ]

        if not moved_ops:
            return op_with_param, op_without_param

        match op_with_param, op_without_param:
            case (Add(), _):
                return (
                    op_or_create_expr(Add, *retained_ops),
                    op_or_create_expr(
                        Add,
                        op_without_param,
                        *[
                            op_or_create_expr(Multiply, op, make_lit(-1))
                            for op in moved_ops
                        ],
                    ),
                )
            case (Multiply(), _):
                return (
                    op_or_create_expr(Multiply, *retained_ops),
                    op_or_create_expr(
                        Multiply,
                        op_without_param,
                        op_or_create_expr(
                            Power, op_or_create_expr(Multiply, *moved_ops), make_lit(-1)
                        ),
                    ),
                )
            case (Power(), _):
                return (
                    op_with_param.operands[0],
                    op_or_create_expr(
                        Power, op_without_param, make_lit(-1)
                    ),  # TODO: fix exponent
                )
            case (_, _):
                return op_with_param, op_without_param

    def isolate_param(
        expr: Expression, param: ParameterOperatable
    ) -> (tuple[ParameterOperatable.All, ParameterOperatable.All]) | None:
        assert len(expr.operands) == 2
        lhs, rhs = expr.operands

        param_in_lhs = param in mutator.utils.find_unique_params(lhs)
        param_in_rhs = param in mutator.utils.find_unique_params(rhs)

        if param_in_lhs and param_in_rhs:
            # TODO
            # supporting this situation means committing to a strategy and sticking to
            # it
            # otherwise we might just make and undo changes until we run out of
            # iterations
            return None

        assert param_in_lhs or param_in_rhs

        op_with_param = lhs if param_in_lhs else rhs
        op_without_param = rhs if param_in_lhs else lhs

        while True:
            new_op_with_param, new_op_without_param = _isolate_param(
                param, op_with_param, op_without_param, from_expr=expr
            )

            if (
                new_op_with_param == op_with_param
                and new_op_without_param == op_without_param
            ):
                break

            op_with_param, op_without_param = new_op_with_param, new_op_without_param

            if op_with_param == param:
                return op_with_param, op_without_param

            # TODO: check for no further progress

    exprs = mutator.nodes_of_type(Is, sort_by_depth=True)
    for expr in exprs:
        if mutator.utils.try_extract_literal(expr) is None:
            continue

        # TODO why? are we trying to do only arithmetic?
        # Then why not do isinstance(expr, Arithmetic)?
        if BoolSet(True) in expr.operands:
            continue

        unaliased_params = {
            p
            for p in mutator.utils.find_unique_params(expr)
            if mutator.utils.try_extract_literal(p) is None
        }

        # only handle single free var
        if len(unaliased_params) != 1:
            continue

        param = unaliased_params.pop()

        if param in expr.operands and not any(
            op is not param and mutator.utils.find_unique_params(op) == {param}
            for op in expr.operands
        ):
            # already isolated
            continue

        if (result := isolate_param(expr, param)) is None:
            continue

        mutator.mutate_expression(expr, operands=result)


@algorithm("Distribute literals across alias classes", terminal=False)
def distribute_literals_across_alias_classes(mutator: Mutator):
    """
    Distribute literals across alias classes

    E is A, A is Lit -> E is Lit
    E is A, A ss Lit -> E ss Lit

    """
    for p in mutator.nodes_of_type():
        lit, is_alias = mutator.utils.try_extract_literal_info(p)
        if lit is None:
            continue

        non_lit_aliases = {
            e: other_p
            for e in p.get_operations(Is, constrained_only=True)
            if not e.get_operand_literals()
            and (other_p := cast_assert(ParameterOperatable, e.get_other_operand(p)))
            is not p
        }
        for alias_expr, alias in non_lit_aliases.items():
            if is_alias:
                mutator.utils.alias_to(alias, lit, from_ops=[p, alias_expr])
            else:
                mutator.utils.subset_to(alias, lit, from_ops=[p, alias_expr])


# Terminal -----------------------------------------------------------------------------


@algorithm("Predicate unconstrained operands deduce", terminal=True)
def predicate_unconstrained_operands_deduce(mutator: Mutator):
    """
    A op! B | A or B unconstrained -> A op!! B
    """

    preds = mutator.nodes_of_type(ConstrainableExpression)
    for p in preds:
        if not p.constrained:
            continue
        if mutator.is_predicate_terminated(p):
            continue
        if mutator.utils.is_literal_expression(p):
            continue

        for op in p.operatable_operands:
            if mutator.utils.no_other_constraints(op, p, unfulfilled_only=True):
                mutator.utils.alias_is_literal_and_check_predicate_eval(p, True)
                break


# Estimation algorithms ----------------------------------------------------------------
@algorithm("Upper estimation", terminal=False)
def upper_estimation_of_expressions_with_subsets(mutator: Mutator):
    """
    If any operand in an expression has a subset literal,
    we can add a subset to the expression.

    ```
    A + B | B is [1,5] -> (A + B) , (A + B) subset (A + [1,5])
    A + B | B subset [1,5] -> (A + B) , (A + B) subset (A + [1,5])
    ```

    No need to check:
    ```
    A + B | A alias B ; never happens (after eq classes)
    A + B | B alias ([0]); never happens (after aliased_single_into_literal)
    ```
    """

    new_literal_mappings = mutator.get_literal_mappings(
        new_only=True, allow_subset=True
    )
    new_exprs = {
        k: v
        for k, v in new_literal_mappings.items()
        if not mutator.utils.is_correlatable_literal(v)
    }

    exprs = {e for alias in new_exprs.keys() for e in alias.get_operations()}
    exprs.update(mutator.non_copy_mutated)
    exprs = ParameterOperatable.sort_by_depth(exprs, ascending=True)

    for expr in exprs:
        assert isinstance(expr, CanonicalExpression)
        # In Is automatically by eq classes
        if isinstance(expr, Is):
            continue
        # Taken care of by singleton fold
        if any(
            mutator.utils.is_replacable_by_literal(op) is not None
            for op in expr.operands
        ):
            continue
        # optimization: don't take away from uncorrelated_alias_fold
        if (
            mutator.terminal
            # not (expr in new_subsets and expr not in new_aliases)
            and not any(mutator.utils.get_correlations(expr))
            and mutator.utils.map_extract_literals(expr)[1]
        ):
            continue
        # In subset useless to look at subset lits
        no_allow_subset_lit = isinstance(expr, IsSubset)

        operands, any_lit = mutator.utils.map_extract_literals(
            expr, allow_subset=not no_allow_subset_lit
        )
        if not any_lit:
            continue

        # TODO make this more efficient (include in extract)
        lit_alias_origins = {
            e
            for p in any_lit
            for e in p.get_operations(Is, constrained_only=True)
            if e.get_operand_literals()
        }

        # Make new expr with subset literals
        mutator.mutate_expression(
            expr,
            operands=operands,
            soft_mutate=IsSubset,
            from_ops=[expr, *lit_alias_origins],
        )


@algorithm("Uncorrelated alias fold", terminal=True)
def uncorrelated_alias_fold(mutator: Mutator):
    """
    If an operation contains only operands that are not correlated with each other,
    we can replace the operands with their corresponding literals.
    ```
    op(As), no correlations in As outside of op, len(A is Lit in As) > 0
    -> op(As) is! op(As replaced by corresponding lits)
    ```

    Terminal because relies on missing correlations.
    """

    new_literal_mappings = mutator.get_literal_mappings(new_only=True)

    # bool expr always map to singles
    new_literal_mappings_filtered = {
        k: v
        for k, v in new_literal_mappings.items()
        if not mutator.utils.is_correlatable_literal(v)
    }
    exprs = {
        e for alias in new_literal_mappings_filtered for e in alias.get_operations()
    }
    # Include mutated since last run
    exprs.update(mutator.non_copy_mutated)
    exprs = ParameterOperatable.sort_by_depth(exprs, ascending=True)

    for expr in exprs:
        assert isinstance(expr, CanonicalExpression)
        # Taken care of by singleton fold
        if any(
            mutator.utils.is_replacable_by_literal(op) is not None
            for op in expr.operands
        ):
            continue
        if isinstance(expr, Is) and expr.constrained:
            # TODO: definitely need to do something
            # just not the same what we do with the other types
            continue
        # TODO: we can weaken this to not replace correlated operands instead of
        #   skipping the whole expression
        # check if any correlations
        if any(mutator.utils.get_correlations(expr)):
            continue

        operands, any_lit = mutator.utils.map_extract_literals(expr)
        if not any_lit:
            continue

        # TODO make this more efficient (include in extract)
        lit_alias_origins = {
            e
            for p in any_lit
            for e in p.get_operations(Is, constrained_only=True)
            if e.get_operand_literals()
        }

        # no point in op! is op! (always true)
        if expr.constrained:
            mutator.create_expression(
                type(expr),
                *operands,
                constrain=True,
                allow_uncorrelated=True,
                from_ops=[expr, *lit_alias_origins],
            )
            continue

        mutator.mutate_expression(
            expr,
            operands=operands,
            soft_mutate=Is,
            from_ops=[expr, *lit_alias_origins],
        )



import logging
from enum import StrEnum
from typing import Callable

from mcp.server.fastmcp import FastMCP
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class Language(StrEnum):
    FABLL = "fabll(python)"
    ATO = "ato"


class NodeType(StrEnum):
    MODULE = "Module"
    INTERFACE = "Interface"


class NodeInfo(BaseModel):
    name: str
    docstring: str
    locator: str
    language: Language
    code: str


class NodeInfoOverview(BaseModel):
    name: str
    docstring: str
    language: Language
    type: NodeType
    inherits: str | None = None


MCP_DECORATOR = Callable[[FastMCP], Callable]


class MCPTools:
    def __init__(self):
        self._tools: dict[Callable, MCP_DECORATOR] = {}

    def register(self, decorator: MCP_DECORATOR = lambda mcp: mcp.tool()):
        def decorator_wrapper(func: Callable):
            self._tools[func] = decorator

            return func

        return decorator_wrapper

    def install(self, mcp: FastMCP):
        for func, decorator in self._tools.items():
            d = decorator(mcp)
            d(func)



import re
from hashlib import sha256


class Checksum:
    class Mismatch(Exception):
        def __init__(self, stated_checksum: str, actual_checksum: str):
            self.stated_checksum = stated_checksum
            self.actual_checksum = actual_checksum

        def __str__(self):
            return f"Checksum mismatch {self.actual_checksum} != {self.stated_checksum}"

    @staticmethod
    def build(to_hash: str, normalize: bool = True) -> str:
        # normalize line endings and whitespace
        if normalize:
            to_hash = to_hash.replace("\r", "")
            c = r" \t"
            s = rf"[{c}]"
            # replace multiple spaces with one space
            to_hash = re.sub(rf"([^{c}\n]){s}+", r"\1 ", to_hash)
            # space before newline
            to_hash = re.sub(rf"{s}+\n", r"\n", to_hash)
            # remove empty lines
            to_hash = re.sub(r"\n+", r"\n", to_hash)

        return sha256(to_hash.encode("utf-8")).hexdigest()

    @staticmethod
    def verify(stated_checksum: str, to_verify: str):
        actual_checksum = Checksum.build(to_verify)

        if (
            actual_checksum != stated_checksum
            # legacy
            and Checksum.build(to_verify, normalize=False) != stated_checksum
        ):
            raise Checksum.Mismatch(stated_checksum, actual_checksum)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module


class can_bridge_by_name(F.can_bridge.impl()):
    def __init__(self, input_name: str = "input", output_name: str = "output"):
        super().__init__()
        self._input_name = input_name
        self._output_name = output_name

    def get_in(self):
        return self.get_obj(Module)[self._input_name]

    def get_out(self):
        return self.get_obj(Module)[self._output_name]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from itertools import pairwise, product

import pytest

import faebryk.library._F as F
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.libs.library import L
from faebryk.libs.test.times import Times
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "count_power, connected",
    product(range(2, 5), [True, False]),
    ids=lambda x: 10 * 2**x
    if not isinstance(x, bool)
    else "connected"
    if x
    else "simple",
)
def test_performance_graph_get_all(count_power: int, connected: bool):
    count = 10 * 2**count_power

    def _factory_simple_resistors():
        class App(Module):
            resistors = L.list_field(count, F.Resistor)

            def __init__(self, timings: Times) -> None:
                super().__init__()
                self._timings = timings

            def __preinit__(self):
                self._timings.add("setup")

        return App

    def _factory_interconnected_resistors():
        class App(Module):
            resistors = L.list_field(count, F.Resistor)

            def __init__(self, timings: Times) -> None:
                super().__init__()
                self._timings = timings

            def __preinit__(self):
                self._timings.add("setup")

                F.Electrical.connect(*(r.unnamed[0] for r in self.resistors))
                self._timings.add("connect")

        return App

    timings = Times()

    if connected:
        AppF = _factory_interconnected_resistors()
    else:
        AppF = _factory_simple_resistors()

    with timings.context("instance"):
        app = AppF(timings)

    G = app.get_graph()
    timings.add("graph")

    G.node_projection()
    timings.add("get_all_nodes_graph")

    for n in [app, app.resistors[0]]:  # type: ignore
        assert isinstance(n, Module)
        name = type(n).__name__[0]

        n.get_children(direct_only=False, types=Node)
        timings.add(f"get_node_children_all {name}")

        n.get_tree(types=Node)
        timings.add(f"get_node_tree {name}")

        n.get_tree(types=ModuleInterface)
        timings.add(f"get_mif_tree {name}")

        n.get_children(direct_only=True, types=Node)
        timings.add(f"get_module_direct_children {name}")

        n.get_children(direct_only=True, types=ModuleInterface)
        timings.add(f"get_mifs {name}")

    logger.info(f"{timings!r}")
    logger.info(str(G))

    per_resistor = timings.get("instance") / count
    logger.info(f"----> Avg/resistor: {per_resistor * 1e6:.2f} us")


def test_performance_graph_merge_rec():
    timings = Times()
    count = 2**14
    logger.info(f"Count: {count}")

    gs = times(count, GraphInterface)
    timings.add("instance")

    def rec_connect(gs_sub: list[GraphInterface]):
        if len(gs_sub) == 1:
            return gs_sub[0]

        mid = len(gs_sub) // 2

        with timings.context(f"split {len(gs_sub)}"):
            timings.add(f"recurse {len(gs_sub)}")
            left = rec_connect(gs_sub[:mid])
            right = rec_connect(gs_sub[mid:])

        timings.add(f"connect {len(gs_sub)}")
        left.connect(right)
        timings.add("connect")

        return left

    with timings.context("total"):
        rec_connect(gs)

    logger.info(timings)
    per_connect = timings.get_formatted("connect", Times.MultiSampleStrategy.AVG)
    logger.info(f"----> Avg/connect: {per_connect}")


def test_performance_graph_merge_it():
    timings = Times(multi_sample_strategy=Times.MultiSampleStrategy.AVG_ACC)
    count = 2**14
    logger.info(f"Count: {count}")

    gs = times(count, GraphInterface)
    timings.add("instance")

    for gl, gr in pairwise(gs):
        gl.connect(gr)
        timings.add("connect")

    assert gs[0].G.node_count == count

    logger.info(timings)
    per_connect = timings.get_formatted("connect", Times.MultiSampleStrategy.AVG)
    logger.info(f"----> Avg/connect: {per_connect}")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from faebryk.core.solver.mutator import Mutator

type SolverAlgorithmFunc = "Callable[[Mutator], None]"


@dataclass(frozen=True)
class SolverAlgorithm:
    name: str
    func: SolverAlgorithmFunc
    single: bool
    terminal: bool

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __str__(self) -> str:
        return self.name


def algorithm(
    name: str,
    single: bool = False,
    terminal: bool = True,
) -> Callable[[SolverAlgorithmFunc], SolverAlgorithm]:
    """
    Decorator to wrap an algorithm function

    Args:
    - single: if True, the algorithm is only applied once in the beginning.
        All other algorithms assume this one ran before
    - terminal: Results are invalid if graph is mutated after solver is run
    """

    if not hasattr(algorithm, "_registered_algorithms"):
        algorithm._registered_algorithms = []

    def decorator(func: SolverAlgorithmFunc) -> SolverAlgorithm:
        @wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        out = SolverAlgorithm(
            name=name,
            func=wrapped,
            single=single,
            terminal=terminal,
        )
        algorithm._registered_algorithms.append(out)

        return out

    return decorator


def get_algorithms() -> list[SolverAlgorithm]:
    return algorithm._registered_algorithms



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import StrEnum

import faebryk.library._F as F
from atopile.errors import UserBadParameterError
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.util import times

logger = logging.getLogger(__name__)


class NetTie(Module):
    """A net tie component that can bridge different interfaces."""

    class PadType(StrEnum):
        SMD = "SMD"
        THT = "THT"

    attach_to_footprint: F.can_attach_to_footprint_symmetrically
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.JP
    )
    picked: F.has_part_removed

    # TODO: generate the kicad footprint instead of loading it
    @L.rt_field
    def footprint(self):
        width_mm = f"{self._width:.1f}mm"
        supported_widths = ["0.5mm", "2.0mm"]
        if width_mm not in supported_widths:
            raise ValueError(
                f"Width [{width_mm}] is currently not supported for NetTie. Supported widths are: {supported_widths}"  # noqa: E501
            )

        if self._pin_count < 2 or self._pin_count > 4:
            raise ValueError(
                f"Pin count [{self._pin_count}] is currently not supported for NetTie"
            )

        fp = F.KicadFootprint(pin_names=[f"{i + 1}" for i in range(self._pin_count)])
        fp.add(
            F.KicadFootprint.has_kicad_identifier(
                f"NetTie:NetTie-{self._pin_count}_{self._pad_type}_Pad{width_mm}"
            )
        )
        return F.has_footprint_defined(fp)

    @L.rt_field
    def power(self):
        return times(self._pin_count, F.ElectricPower)

    def __init__(
        self,
        width: float = 0.5,
        pin_count: int = 2,
        pad_type: PadType = PadType.SMD,
        connect_gnd: bool = True,
    ) -> None:
        super().__init__()
        self._width = width
        self._pin_count = pin_count
        self._pad_type = pad_type
        self._connect_gnd = connect_gnd

    def __preinit__(self):
        if self._pin_count == 1:
            raise UserBadParameterError("NetTie with <2 pins is not allowed")

        if self._pin_count == 2:
            self.add(F.can_bridge_defined(*self.power))

        # Connect all interfaces to the first one
        for p in self.power[1:]:
            p.connect_shallow(self.power[0])

        # connect to footprint
        self.footprint.get_footprint().get_trait(F.can_attach_via_pinmap).attach(
            pinmap={
                f"{i + 1}": power.lv if self._connect_gnd else power.hv
                for i, power in enumerate(self.power)
            }
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

# TODO this used to be interactive_graph.py
# merge it back into it

from itertools import pairwise
from typing import Any, Callable, Collection, Iterable

import dash_core_components as dcc
import dash_cytoscape as cyto
from dash import Dash, html
from dash.dependencies import Input, Output, State
from rich.console import Console
from rich.table import Table

# import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.graphinterface import GraphInterface
from faebryk.core.link import Link, LinkSibling
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.core.parameter import Expression, Parameter, Predicate
from faebryk.core.trait import Trait
from faebryk.exporters.visualize.util import generate_pastel_palette
from faebryk.libs.util import KeyErrorAmbiguous, cast_assert, find_or, groupby, typename


# Transformers -------------------------------------------------------------------------
def _gif(gif: GraphInterface):
    return {
        "data": {
            "id": str(id(gif)),
            "label": gif.name,
            "type": typename(gif),
            "parent": str(id(gif.node)),
        }
    }


def _link(source, target, link: Link):
    return {
        "data": {
            "source": str(id(source)),
            "target": str(id(target)),
            "type": typename(link),
        }
    }


_GROUP_TYPES = {
    Predicate: "#FCF3CF",  # Very light goldenrod
    Expression: "#D1F2EB",  # Very soft turquoise
    Parameter: "#FFD9DE",  # Very light pink
    Module: "#E0F0FF",  # Very light blue
    Trait: "#FCFCFF",  # Almost white
    # F.Electrical: "#D1F2EB",  # Very soft turquoise
    # F.ElectricPower: "#FCF3CF",  # Very light goldenrod
    # F.ElectricLogic: "#EBE1F1",  # Very soft lavender
    # Defaults
    ModuleInterface: "#DFFFE4",  # Very light green
    Node: "#FCFCFF",  # Almost white
}


def _group(node: Node, root: bool):
    try:
        subtype = find_or(_GROUP_TYPES, lambda t: isinstance(node, t), default=Node)
    except KeyErrorAmbiguous as e:
        subtype = e.duplicates[0]

    if root:
        hier = node.get_hierarchy()
        type_hier = [t for t, _ in hier]
        name_hier = [n for _, n in hier]
        name = ".".join(name_hier)
        types = "|".join(typename(t) for t in type_hier)
        label = f"{name}\n({types})"
    else:
        label = f"{node.get_name(accept_no_parent=True)}\n({typename(node)})"

    return {
        "data": {
            "id": str(id(node)),
            "label": label,
            "type": "group",
            "subtype": typename(subtype),
            "parent": str(id(p[0])) if (p := node.get_parent()) else None,
        }
    }


# Style --------------------------------------------------------------------------------


def _with_pastels(iterable: Collection[str]):
    return zip(sorted(iterable), generate_pastel_palette(len(iterable)))


class _Stylesheet:
    _BASE = [
        {
            "selector": "node",
            "style": {
                "content": "data(label)",
                "text-opacity": 0.8,
                "text-valign": "center",
                "text-halign": "center",
                "font-size": "0.5em",
                "background-color": "#BFD7B5",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 0.5,
                "border-width": 1,
                "border-color": "#888888",
                "border-opacity": 0.5,
            },
        },
        {
            "selector": "edge",
            "style": {
                "width": 1,
                "line-color": "#A3C4BC",
                "curve-style": "bezier",
                "target-arrow-shape": "triangle",
                "arrow-scale": 1,
                "target-arrow-color": "#A3C4BC",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 2,
            },
        },
        {
            "selector": 'node[type = "group"]',
            "style": {
                "background-color": "#D3D3D3",
                "font-weight": "bold",
                "font-size": "1.5em",
                "text-valign": "top",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 1.5,
                "text-wrap": "wrap",
                "border-width": 4,
            },
        },
    ]

    def __init__(self):
        self.stylesheet = list(self._BASE)

    def add_node_type(self, node_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'node[type = "{node_type}"]',
                "style": {"background-color": color},
            }
        )

    def add_link_type(self, link_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'edge[type = "{link_type}"]',
                "style": {
                    "line-color": color,
                    "target-arrow-color": color,
                    # "target-arrow-shape": "none",
                    # "source-arrow-shape": "none",
                },
            }
        )

    def add_group_type(self, group_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'node[subtype = "{group_type}"]',
                "style": {"background-color": color},
            }
        )


def _Layout(
    stylesheet: _Stylesheet, elements: list[dict[str, dict]], extra: dict | None = None
):
    if not extra:
        extra = {}
    return html.Div(
        style={
            "position": "fixed",
            "display": "flex",
            "flex-direction": "column",
            "height": "100%",
            "width": "100%",
        },
        children=[
            html.Div(
                className="cy-container",
                style={"flex": "1", "position": "relative"},
                children=[
                    cyto.Cytoscape(
                        id="graph-view",
                        stylesheet=stylesheet.stylesheet,
                        style={
                            "position": "absolute",
                            "width": "100%",
                            "height": "100%",
                            "zIndex": 999,
                        },
                        elements=elements,
                        layout={
                            "name": "fcose",
                        }
                        | extra,
                    )
                ],
            ),
        ],
    )


def _get_layout(app: Dash) -> dict[str, Any]:
    for html_node in cast_assert(list, cast_assert(html.Div, app.layout).children):
        if not isinstance(html_node, html.Div):
            continue
        for child in cast_assert(list, html_node.children):
            if not isinstance(child, cyto.Cytoscape):
                continue
            return child.layout
    raise ValueError("No Cytoscape found in layout")


# --------------------------------------------------------------------------------------


class Layout:
    type ID_or_OBJECT = object | str

    def __init__(self, app: Dash, elements: list[dict], nodes: list[Node]):
        self.app = app
        self.layout = _get_layout(app)
        self.ids = {
            element["data"]["id"] for element in elements if "id" in element["data"]
        }

        self.div_children = cast_assert(
            list, cast_assert(html.Div, app.layout).children
        )
        self.nodes = nodes

    def is_filtered(self, elem: ID_or_OBJECT) -> bool:
        if not isinstance(elem, str):
            elem = self.id_of(elem)
        return elem not in self.ids

    def nodes_of_type[T: Node](self, node_type: type[T]) -> set[T]:
        return {
            n
            for n in self.nodes
            if isinstance(n, node_type) and not self.is_filtered(n.self_gif)
        }

    @staticmethod
    def id_of(obj: ID_or_OBJECT) -> str:
        if isinstance(obj, str):
            return obj
        return str(id(obj))

    def add_rel_constraint(
        self,
        source: ID_or_OBJECT,
        target: ID_or_OBJECT,
        gap_x: int | None = None,
        gap_y: int | None = None,
        layout: dict | None = None,
    ):
        if not layout:
            layout = self.layout

        if "relativePlacementConstraint" not in layout:
            layout["relativePlacementConstraint"] = []
        rel_placement_constraints = cast_assert(
            list, layout["relativePlacementConstraint"]
        )

        if self.is_filtered(source) or self.is_filtered(target):
            return
        if gap_y is not None:
            top, bot = (source, target) if gap_y >= 0 else (target, source)

            # if isinstance(top, GraphInterface) and isinstance(bot, GraphInterface):
            #     print(f"{top}\n   v\n{bot}")

            rel_placement_constraints.append(
                {
                    "top": self.id_of(top),
                    "bottom": self.id_of(bot),
                    "gap": abs(gap_y),
                }
            )
        if gap_x is not None:
            left, right = (source, target) if gap_x >= 0 else (target, source)

            # if isinstance(left, GraphInterface) and isinstance(right, GraphInterface):
            #    print(f"{left} > {right}")

            rel_placement_constraints.append(
                {
                    "left": self.id_of(left),
                    "right": self.id_of(right),
                    "gap": abs(gap_x),
                }
            )

    def add_rel_top_bot(
        self,
        top: ID_or_OBJECT,
        bot: ID_or_OBJECT,
        gap: int = 0,
        layout: dict | None = None,
    ):
        assert gap >= 0
        self.add_rel_constraint(top, bot, gap_y=gap, layout=layout)

    def add_rel_left_right(
        self,
        left: ID_or_OBJECT,
        right: ID_or_OBJECT,
        gap: int = 0,
        layout: dict | None = None,
    ):
        assert gap >= 0
        self.add_rel_constraint(left, right, gap_x=gap, layout=layout)

    def add_order(
        self,
        *nodes: ID_or_OBJECT,
        horizontal: bool,
        gap: int = 0,
        layout: dict | None = None,
    ):
        if not layout:
            layout = self.layout
        for n1, n2 in pairwise(nodes):
            if horizontal:
                self.add_rel_left_right(n1, n2, gap=gap, layout=layout)
            else:
                self.add_rel_top_bot(n1, n2, gap=gap, layout=layout)

    def add_align(
        self, *nodes: ID_or_OBJECT, horizontal: bool, layout: dict | None = None
    ):
        if not layout:
            layout = self.layout
        direction = "horizontal" if horizontal else "vertical"
        nodes = tuple(n for n in nodes if not self.is_filtered(n))
        if len(nodes) <= 1:
            return

        # if all(isinstance(n, GraphInterface) for n in nodes):
        #     print(f"align {direction}: {nodes}")

        if "alignmentConstraint" not in layout:
            layout["alignmentConstraint"] = {}
        if direction not in layout["alignmentConstraint"]:
            layout["alignmentConstraint"][direction] = []

        align = cast_assert(dict, layout["alignmentConstraint"])
        align[direction].append([self.id_of(n) for n in nodes])

    def add_same_height[T: Node](
        self,
        nodes: Iterable[T],
        gif_key: Callable[[T], GraphInterface],
        layout: dict | None = None,
    ):
        if not layout:
            layout = self.layout
        self.add_align(*(gif_key(n) for n in nodes), horizontal=True, layout=layout)

    def set_type(self, t: str, layout: dict | None = None):
        if not layout:
            layout = self.layout
        if t == "fcose" or t is None:
            _layout = {
                "name": "fcose",
                # 'draft', 'default' or 'proof'
                # - "draft" only applies spectral layout
                # - "default" improves the quality with incremental layout
                #   (fast cooling rate)
                # - "proof" improves the quality with incremental layout
                #   (slow cooling rate)
                "quality": "proof",
                # Whether or not to animate the layout
                "animate": False,
                # Use random node positions at beginning of layout
                # if this is set to false,
                # then quality option must be "proof"
                "randomize": False,
                # Fit the viewport to the repositioned nodes
                "fit": True,
                # Padding around layout
                "padding": 50,
                # Whether to include labels in node dimensions.
                # Valid in "proof" quality
                "nodeDimensionsIncludeLabels": True,
                # Whether or not simple nodes (non-compound nodes)
                #  are of uniform dimensions
                "uniformNodeDimensions": True,
                # Whether to pack disconnected components -
                # cytoscape-layout-utilities extension should
                # be registered and initialized
                "packComponents": False,  # Graph is never disconnected
                # Node repulsion (non overlapping) multiplier
                "nodeRepulsion": 100,
                # Ideal edge (non nested) length
                "idealEdgeLength": 100,
                # Divisor to compute edge forces
                "edgeElasticity": 0.2,
                # Nesting factor (multiplier) to compute ideal edge length
                # for nested edges
                "nestingFactor": 0.0001,
                # Maximum number of iterations to perform -
                # this is a suggested value and might be adjusted by the
                #  algorithm as required
                "numIter": 2500 * 4,
                # For enabling tiling
                "tile": False,  # No unconnected nodes in Graph
                # Gravity force (constant)
                "gravity": 0,
                # Gravity range (constant)
                "gravityRange": 3.8,
                # Gravity force (constant) for compounds
                "gravityCompound": 20,
                # Gravity range (constant) for compounds
                "gravityRangeCompound": 0.5,
                # Initial cooling factor for incremental layout
                "initialEnergyOnIncremental": 0.5,
                "componentSpacing": 40,
            }
        elif t == "dagre":
            _layout = {
                "name": "dagre",
            }
        else:
            raise ValueError(f"Unknown layout: {t}")

        layout.clear()
        layout.update(_layout)


def buttons(layout: Layout):
    app = layout.app
    html_controls = html.Div(
        className="controls",
        style={"padding": "10px", "background-color": "#f0f0f0"},
        children=[
            #         html.Label("Node Repulsion:"),
            #         dcc.Slider(
            #             id="node-repulsion-slider",
            #             min=500,
            #             max=5000,
            #             step=100,
            #             value=1000,
            #             marks={i: str(i) for i in range(500, 5001, 500)},
            #         ),
            #         html.Label("Edge Elasticity:"),
            #         dcc.Slider(
            #             id="edge-elasticity-slider",
            #             min=0,
            #             max=1,
            #             step=0.05,
            #             value=0.45,
            #             marks={i / 10: str(i / 10) for i in range(0, 11, 1)},
            #         ),
            dcc.RadioItems(
                id="layout-radio",
                options=[
                    {"label": "fcose", "value": "fcose"},
                    {"label": "dagre", "value": "dagre"},
                ],
            ),
            dcc.RadioItems(
                id="layout-dagre-ranker",
                options=[
                    {"label": "network-simplex", "value": "network-simplex"},
                    {"label": "tight-tree", "value": "tight-tree"},
                    {"label": "longest-path", "value": "longest-path"},
                ],
            ),
            dcc.Checklist(
                id="layout-checkbox",
                options=[{"label": "Parameters", "value": "parameters"}],
            ),
            html.Button("Apply Changes", id="apply-changes-button"),
        ],
    )
    layout.div_children.insert(-2, html_controls)

    @app.callback(
        Output("graph-view", "layout"),
        Input("apply-changes-button", "n_clicks"),
        State("layout-checkbox", "value"),
        State("layout-radio", "value"),
        State("layout-dagre-ranker", "value"),
        State("graph-view", "layout"),
    )
    def absolute_layout(
        n_clicks, layout_checkbox, layout_radio, layout_dagre_ranker, current_layout
    ):
        print(layout_checkbox, layout_radio, layout_dagre_ranker)
        layout.set_type(layout_radio, current_layout)

        if layout_radio == "fcose":
            layout_constraints(layout, current_layout)

        if "parameters" in (layout_checkbox or []):
            params_top(layout, current_layout)

        if layout_dagre_ranker:
            current_layout["ranker"] = layout_dagre_ranker

        return current_layout


def params_top(layout: Layout, _layout: dict | None = None):
    params = layout.nodes_of_type(Parameter)
    expressions = layout.nodes_of_type(Expression)
    predicates = layout.nodes_of_type(Predicate)
    non_predicate_expressions = expressions - predicates

    def depth(expr: Expression) -> int:
        operates_on = expressions & {
            e.node
            for e, li in expr.operates_on.edges.items()
            if not isinstance(li, LinkSibling) and e.node is not expr
        }

        # direct parameter or constants only
        if not operates_on:
            return 1
        return 1 + max(depth(o) for o in operates_on)

    expressions_by_depth = groupby(non_predicate_expressions, depth)

    def same_height[T: Parameter | Expression](nodes: Iterable[T]):
        layout.add_same_height(nodes, lambda pe: pe.self_gif, layout=_layout)
        layout.add_same_height(nodes, lambda pe: pe.operated_on, layout=_layout)

    # All params same height
    same_height(params)

    # predicates same height
    same_height(predicates)

    for _, exprs in expressions_by_depth.items():
        same_height(exprs)

    # predicate expressions below other expressions
    if predicates:
        for expr in non_predicate_expressions:
            layout.add_rel_top_bot(
                expr.operates_on, next(iter(predicates)).self_gif, gap=200
            )

    # Expressions below params
    if params:
        for expr in expressions:
            layout.add_rel_top_bot(
                next(iter(params)).operated_on, expr.self_gif, gap=200
            )

    # Expression tree
    for expr in expressions:
        operates_on = (params | expressions) & {
            e.node
            for e, li in expr.operates_on.edges.items()
            if not isinstance(li, LinkSibling) and e.node is not expr
        }
        for o in operates_on:
            layout.add_rel_top_bot(o.operated_on, expr.self_gif, gap=100)


def layout_constraints(layout: Layout, _layout: dict | None = None):
    for n in layout.nodes:
        # only to save on some printing
        if layout.is_filtered(n.self_gif):
            continue

        siblings = {
            o
            for o, li in n.self_gif.edges.items()
            if isinstance(li, LinkSibling) and not layout.is_filtered(o)
        }

        # siblings below self
        for o in siblings:
            layout.add_rel_top_bot(n.self_gif, o, gap=50, layout=_layout)

        # siblings on same level within node
        layout.add_align(*siblings, horizontal=True, layout=_layout)

        order = list(sorted(siblings, key=lambda o: o.name))
        middle_i = len(order) // 2
        if len(siblings) % 2 == 1:
            # sibling directly below self
            layout.add_align(
                n.self_gif, order[middle_i], horizontal=False, layout=_layout
            )
            order.pop(middle_i)

        # self inbetween siblings
        order.insert(middle_i, n.self_gif)
        layout.add_order(*order, horizontal=True, gap=25, layout=_layout)


# --------------------------------------------------------------------------------------


def interactive_subgraph(
    edges: Iterable[tuple[GraphInterface, GraphInterface, Link]],
    gifs: list[GraphInterface],
    nodes: Iterable[Node],
    height: int | None = None,
):
    links = [link for _, _, link in edges]
    link_types = {typename(link) for link in links}
    gif_types = {typename(gif) for gif in gifs}

    def node_has_parent_in_graph(node: Node) -> bool:
        p = node.get_parent()
        if not p:
            return False
        return p[0] in nodes

    elements = (
        [_gif(gif) for gif in gifs]
        + [_link(*edge) for edge in edges]
        + [_group(node, root=not node_has_parent_in_graph(node)) for node in nodes]
    )

    # Build stylesheet
    stylesheet = _Stylesheet()

    gif_type_colors = list(_with_pastels(gif_types))
    link_type_colors = list(_with_pastels(link_types))
    group_types_colors = [
        (typename(group_type), color) for group_type, color in _GROUP_TYPES.items()
    ]

    for gif_type, color in gif_type_colors:
        stylesheet.add_node_type(gif_type, color)

    for link_type, color in link_type_colors:
        stylesheet.add_link_type(link_type, color)

    for group_type, color in group_types_colors:
        stylesheet.add_group_type(group_type, color)

    # Register the fcose layout
    cyto.load_extra_layouts()
    app = Dash(__name__)
    app.layout = _Layout(stylesheet, elements)

    # Extra layouting
    layout = Layout(app, elements, list(nodes))
    buttons(layout)

    # Print legend ---------------------------------------------------------------------
    console = Console()

    for typegroup, colors in [
        ("GIF", gif_type_colors),
        ("Link", link_type_colors),
        ("Node", group_types_colors),
    ]:
        table = Table(title="Legend")
        table.add_column("Type", style="cyan")
        table.add_column("Color", style="green")
        table.add_column("Name")

        for text, color in colors:
            table.add_row(typegroup, f"[on {color}]    [/]", text)

        console.print(table)

    # Run ------------------------------------------------------------------------------

    app.run(jupyter_height=height or 1000)


def interactive_graph(
    G: Graph,
    node_types: tuple[type[Node], ...] | None = None,
    depth: int = 0,
    filter_unconnected: bool = True,
    height: int | None = None,
):
    if node_types is None:
        node_types = (Node,)

    # Build elements
    nodes = GraphFunctions(G).nodes_of_types(node_types)
    if depth > 0:
        nodes = [node for node in nodes if len(node.get_hierarchy()) <= depth]

    gifs = {gif for gif in G.get_gifs() if gif.node in nodes}
    if filter_unconnected:
        gifs = [gif for gif in gifs if len(gif.edges.keys() & gifs) > 1]

    edges = [
        (edge[0], edge[1], edge[2])
        for edge in G.edges
        if edge[0] in gifs and edge[1] in gifs
    ]
    return interactive_subgraph(edges, list(gifs), nodes, height=height)



import rich

import faebryk.libs.logging

rich.reconfigure(theme=faebryk.libs.logging.theme)

# Console should be a singleton to avoid intermixing logging w/ other output
console = rich.get_console()
error_console = rich.console.Console(theme=faebryk.libs.logging.theme, stderr=True)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.libs.library import L  # noqa: F401


class Logic(F.Signal):
    """
    Acts as protocol, because multi inheritance is not supported
    """

    # state = L.p_field(domain=L.Domains.BOOL())

    def set(self, on: bool):
        # self.state.constrain_subset(on)
        ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class MOSFET(Module):
    class ChannelType(Enum):
        N_CHANNEL = auto()
        P_CHANNEL = auto()

    class SaturationType(Enum):
        ENHANCEMENT = auto()
        DEPLETION = auto()

    channel_type = L.p_field(domain=L.Domains.ENUM(ChannelType))
    saturation_type = L.p_field(domain=L.Domains.ENUM(SaturationType))
    gate_source_threshold_voltage = L.p_field(units=P.V)
    max_drain_source_voltage = L.p_field(units=P.V)
    max_continuous_drain_current = L.p_field(units=P.A)
    on_resistance = L.p_field(units=P.ohm)

    source: F.Electrical
    gate: F.Electrical
    drain: F.Electrical

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.Q
    )

    # @L.rt_field
    # def pickable(self) -> F.is_pickable_by_type:
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.MOSFET,
    #         {
    #             "channel_type": self.channel_type,
    #             # TODO: add support in backend
    #             # "saturation_type": self.saturation_type,
    #             "gate_source_threshold_voltage": self.gate_source_threshold_voltage,
    #             "max_drain_source_voltage": self.max_drain_source_voltage,
    #             "max_continuous_drain_current": self.max_continuous_drain_current,
    #             "on_resistance": self.on_resistance,
    #         },
    #     )

    # TODO pretty confusing
    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(in_if=self.source, out_if=self.drain)

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.source: ["S", "Source"],
                self.gate: ["G", "Gate"],
                self.drain: ["D", "Drain"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from abc import abstractmethod

import pytest


def test_add():
    from faebryk.core.cpp import add, call_python_function

    assert add(1, 2) == 3
    assert add(1) == 2

    assert call_python_function(lambda: 1) == 1


def test_cnodes():
    from faebryk.core.cpp import LinkNamedParent, Node

    n1 = Node()
    n1.transfer_ownership(n1)
    n2 = Node()
    n2.transfer_ownership(n2)

    class _Node(Node):
        def __init__(self) -> None:
            super().__init__()
            self.transfer_ownership(self)

    n3 = _Node()

    n1.children.connect(n2.parent, LinkNamedParent("test1"))
    n2.children.connect(n3.parent, LinkNamedParent("test2"))
    print(n1)
    print(n2)
    print(n3)
    print(n1.children.get_children())
    print(n1.get_children(direct_only=True, sort=True))
    print(n2.get_children(direct_only=True, sort=True))
    print(n1.get_children(include_root=True, direct_only=False, sort=True))


def test_pynode():
    from faebryk.core.node import Node

    n = Node()
    print(n)
    print("---")

    class SubNode(Node):
        a: Node
        b: Node

    sn = SubNode()
    print(sn.a)

    print(sn.get_children(direct_only=True, types=Node))


def test_derived_pynodes():
    from faebryk.core.module import Module
    from faebryk.core.moduleinterface import ModuleInterface

    class App(Module):
        mif1: ModuleInterface
        mif2: ModuleInterface

    app = App()
    app.mif1.connect(app.mif2)

    print(app.mif1)
    print(app.mif1.get_connected())


def test_traits_basic():
    from faebryk.core.node import Node
    from faebryk.core.trait import Trait

    class T(Trait):
        @abstractmethod
        def do(self): ...

    class T_do(T.impl()):
        def do(self):
            print("do")

    class A(Node):
        t: T_do

    a = A()

    print(a.t)
    print(a.get_trait(T))
    a.get_trait(T).do()


def test_forgotten_superinit():
    from faebryk.core.node import Node

    class A(Node):
        def __init__(self):
            pass

    with pytest.raises(Exception):
        A()


def test_library_nodes():
    import faebryk.library._F as F

    x = F.Electrical()

    print(x)


def test_cobject():
    from faebryk.core.cpp import (
        GraphInterface,
        GraphInterfaceHierarchical,
        GraphInterfaceSelf,
        LinkDirect,
        LinkNamedParent,
    )

    g1 = GraphInterfaceSelf()
    g2 = GraphInterfaceHierarchical(True)
    g3 = GraphInterface()
    g4 = GraphInterfaceHierarchical(False)

    g2.connect(g3)
    g1.connect(g2, LinkDirect())
    g2.connect(g4, LinkNamedParent("test"))

    print(g1.edges)
    print(g2.edges)
    print(g1.get_graph())

    g1.get_graph().invalidate()


def test_link():
    from faebryk.core.cpp import GraphInterface, LinkDirect

    g1 = GraphInterface()
    g2 = GraphInterface()
    g1.connect(g2, LinkDirect())


def test_mif_link():
    from faebryk.core.link import LinkDirectConditional
    from faebryk.core.moduleinterface import ModuleInterface

    mif1 = ModuleInterface()
    mif2 = ModuleInterface()
    mif1.connect_shallow(mif2)
    paths = mif1.is_connected_to(mif2)
    assert len(paths) == 1
    path = paths[0]
    assert len(path) == 4
    assert isinstance(path[1].is_connected_to(path[2]), LinkDirectConditional)


def test_cpp_type():
    from faebryk.core.cpp import LinkDirect

    class LinkDirectDerived(LinkDirect):
        pass

    obj = LinkDirect()
    assert LinkDirect.is_cloneable(obj)

    obj2 = LinkDirectDerived()
    assert not LinkDirect.is_cloneable(obj2)


def test_isinstance_base():
    from faebryk.core.node import Node

    assert Node._mro == []
    assert Node._mro_ids == set()

    n = Node()
    assert n.isinstance(Node)
    assert n.isinstance([Node])


def test_isinstance_existing():
    from faebryk.core.module import Module
    from faebryk.core.node import Node

    assert Module._mro == [Module, Node]
    assert Module._mro_ids == {id(Module), id(Node)}

    m = Module()
    assert m.isinstance(Module)
    assert m.isinstance(Node)


def test_isinstance_new():
    from faebryk.core.module import Module
    from faebryk.core.node import Node

    class A(Module):
        pass

    class B(Module):
        pass

    a = A()
    assert a.isinstance(A)
    assert a.isinstance(Module)
    assert a.isinstance(Node)
    assert not a.isinstance(B)

    assert a.isinstance([A, Module, Node])
    assert a.isinstance([B, Module])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from enum import Enum, auto

from deprecated import deprecated

import faebryk.library._F as F
from faebryk.core.parameter import ParameterOperatable
from faebryk.libs.library import L
from faebryk.libs.units import P


class LED(F.Diode):
    class Color(Enum):
        # Primary Colors
        RED = auto()
        GREEN = auto()
        BLUE = auto()

        # Secondary and Mixed Colors
        YELLOW = auto()
        ORANGE = auto()
        PURPLE = auto()
        CYAN = auto()
        MAGENTA = auto()

        # Shades of White
        WHITE = auto()
        WARM_WHITE = auto()
        COLD_WHITE = auto()
        NATURAL_WHITE = auto()

        # Other Colors
        EMERALD = auto()
        AMBER = auto()
        PINK = auto()
        LIME = auto()
        VIOLET = auto()

        # Specific LED Colors
        ULTRA_VIOLET = auto()
        INFRA_RED = auto()

    brightness = L.p_field(units=P.candela)
    max_brightness = L.p_field(units=P.candela)
    color = L.p_field(domain=L.Domains.ENUM(Color))

    # @L.rt_field
    # def pickable(self):
    #     return F.is_pickable_by_type(
    #         F.is_pickable_by_type.Type.LED,
    #         {
    #             # Diode
    #             "forward_voltage": self.forward_voltage,
    #             "reverse_working_voltage": self.reverse_working_voltage,
    #             "reverse_leakage_current": self.reverse_leakage_current,
    #             # LED
    #             "max_current": self.max_current,
    #             "max_brightness": self.max_brightness,
    #             "color": self.color,
    #         },
    #     )

    def __preinit__(self):
        self.current.alias_is(self.brightness / self.max_brightness * self.max_current)
        self.brightness.constrain_le(self.max_brightness)

    def set_intensity(self, intensity: ParameterOperatable.NumberLike) -> None:
        self.brightness.alias_is(intensity * self.max_brightness)

    @deprecated(reason="Use PoweredLED instead")
    def connect_via_current_limiting_resistor(
        self,
        input_voltage: ParameterOperatable.NumberLike,
        resistor: F.Resistor,
        target: F.Electrical,
        low_side: bool,
    ):
        if low_side:
            self.cathode.connect_via(resistor, target)
        else:
            self.anode.connect_via(resistor, target)

        resistor.resistance.alias_is(
            self.get_needed_series_resistance_for_current_limit(input_voltage),
        )
        resistor.allow_removal_if_zero()

    @deprecated(reason="Use PoweredLED instead")
    def connect_via_current_limiting_resistor_to_power(
        self, resistor: F.Resistor, power: F.ElectricPower, low_side: bool
    ):
        if low_side:
            self.anode.connect(power.hv)
        else:
            self.cathode.connect(power.lv)

        self.connect_via_current_limiting_resistor(
            power.voltage,
            resistor,
            power.lv if low_side else power.hv,
            low_side,
        )



import sys

# fast-path for self-check
# makes extension a lot faster
if __name__ in ("__main__", "atopile.cli.cli"):
    if len(sys.argv) == 2 and sys.argv[1] == "self-check":
        from importlib.metadata import version as get_package_version

        print(get_package_version("atopile"))
        sys.exit(0)


# excepthook must be installed before typer is imported
import atopile.cli.excepthook  # noqa: F401, I001

import json
import logging
from enum import Enum
from importlib.metadata import version as get_package_version
from pathlib import Path
from typing import Annotated

import typer

from atopile import version
from atopile.cli import (
    build,
    configure,
    create,
    inspect_,
    install,
    package,
    view,
    lsp,
    mcp,
)
from atopile.cli.logging_ import handler, logger
from atopile.config import config
from atopile.errors import UserException, UserNoProjectException
from atopile.version import check_for_update
from faebryk.libs.exceptions import (
    UserResourceException,
    iter_leaf_exceptions,
)
from faebryk.libs.logging import FLOG_FMT

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=bool(FLOG_FMT),  # required to override the excepthook
    rich_markup_mode="rich",
)


def python_interpreter_path(ctx: typer.Context, value: bool):
    """Print the current python interpreter path."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(sys.executable)
    raise typer.Exit()


def atopile_src_path(ctx: typer.Context, value: bool):
    """Print the current python interpreter path."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(Path(__file__).parent.parent)
    raise typer.Exit()


def version_callback(ctx: typer.Context, value: bool):
    """Output a version string meeting the pypa version spec."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(get_package_version("atopile"))
    raise typer.Exit()


def semver_callback(ctx: typer.Context, value: bool):
    """Output a version string meeting the semver.org spec."""
    if not value or ctx.resilient_parsing:
        return
    version_string = get_package_version("atopile")
    typer.echo(version.parse(version_string))
    raise typer.Exit()


@app.callback()
def cli(
    ctx: typer.Context,
    non_interactive: Annotated[
        bool | None,
        typer.Option(
            "--non-interactive", envvar=["ATO_NON_INTERACTIVE", "NONINTERACTIVE"]
        ),
    ] = None,
    debug: Annotated[
        bool,
        typer.Option("--debug", help="Wait to attach debugger on start"),
    ] = False,
    verbose: Annotated[
        int,
        typer.Option("--verbose", "-v", count=True, help="Increase verbosity"),
    ] = 0,
    python_path: Annotated[
        bool, typer.Option(hidden=True, callback=python_interpreter_path)
    ] = False,
    atopile_path: Annotated[
        bool, typer.Option(hidden=True, callback=atopile_src_path)
    ] = False,
    version: Annotated[
        bool | None,
        typer.Option("--version", callback=version_callback, is_eager=True),
    ] = None,
    semver: Annotated[
        bool | None,
        typer.Option("--semver", callback=semver_callback, is_eager=True),
    ] = None,
):
    if debug:
        import debugpy  # pylint: disable=import-outside-toplevel

        debug_port = 5678
        debugpy.listen(("localhost", debug_port))
        logger.info("Starting debugpy on port %s", debug_port)
        debugpy.wait_for_client()

    # set the log level
    if verbose == 1:
        handler.hide_traceback_types = ()
        handler.tracebacks_show_locals = True
    elif verbose == 2:
        handler.tracebacks_suppress_map = {}  # Traceback through atopile infra
    elif verbose >= 3:
        logger.root.setLevel(logging.DEBUG)
        handler.traceback_level = logging.WARNING

    # FIXME: this won't work properly when configs
    # are reloaded from a pointed-to file (eg in `ato build path/to/file`)
    # from outside a project directory
    if non_interactive is not None:
        config.interactive = not non_interactive

    if ctx.invoked_subcommand:
        check_for_update()

    if config.interactive and ctx.invoked_subcommand != "configure":
        configure.do_configure_if_needed()


app.command()(build.build)
app.add_typer(create.create_app, name="create")
app.command(deprecated=True, hidden=True)(install.install)
app.command()(configure.configure)
app.command()(inspect_.inspect)
app.command()(view.view)
app.add_typer(package.package_app, name="package", hidden=True)
app.add_typer(install.dependencies_app, name="dependencies", help="Manage dependencies")
app.command(rich_help_panel="Shortcuts")(install.sync)
app.command(rich_help_panel="Shortcuts")(install.add)
app.command(rich_help_panel="Shortcuts")(install.remove)
app.add_typer(lsp.lsp_app, name="lsp", hidden=True)
app.add_typer(mcp.mcp_app, name="mcp", hidden=True)


@app.command(hidden=True)
def export_config_schema(pretty: bool = False):
    from atopile.config import ProjectConfig

    config_schema = ProjectConfig.model_json_schema()

    if pretty:
        print(json.dumps(config_schema, indent=4))
    else:
        print(json.dumps(config_schema))


class ConfigFormat(str, Enum):
    python = "python"
    json = "json"


@app.command(hidden=True)
def dump_config(format: ConfigFormat = ConfigFormat.python):
    from rich import print

    print(config.project.model_dump(mode=format))


@app.command(help="Check file for syntax errors and internal consistency")
def validate(
    path: Annotated[Path, typer.Argument(exists=True, file_okay=True, dir_okay=False)],
):
    from atopile import front_end

    path = path.resolve().relative_to(Path.cwd())

    # pick up project config if we're in a project
    # required for package search path inclusion
    try:
        config.apply_options(entry=None)
    except UserNoProjectException:
        pass

    if path.suffix != ".ato":
        raise UserResourceException("Invalid file type")

    try:
        front_end.bob.try_build_all_from_file(path)
    except* UserException as e:
        for error in iter_leaf_exceptions(e):
            logger.error(error, exc_info=error)

    else:
        typer.echo(f"{path}: ok")


def main():
    app()


if __name__ == "__main__":
    main()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.core.node import Node
from faebryk.core.trait import Trait


class Footprint(Module):
    class TraitT(Trait): ...

    @staticmethod
    def get_footprint_of_parent(
        intf: ModuleInterface,
    ) -> "tuple[Node, Footprint]":
        parent, trait = intf.get_parent_with_trait(F.has_footprint)
        return parent, trait.get_footprint()



#!/usr/bin/env python
import pathlib

if __name__ == "__main__":
    if "Not open source" == "{{ cookiecutter.license }}":
        try:
            pathlib.Path("LICENSE").unlink()
        except FileNotFoundError:
            pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from pathlib import Path
from typing import cast

import dash_core_components as dcc
import dash_cytoscape as cyto
from dash import Dash, html
from dash.dependencies import Input, Output, State

# import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.link import LinkSibling
from faebryk.core.node import Node
from faebryk.core.parameter import Expression, Parameter
from faebryk.exporters.parameters.parameters_to_file import parameter_report
from faebryk.exporters.visualize.interactive_params_base import (
    _GROUP_TYPES,
    Layout,
    _Layout,
)
from faebryk.libs.util import (
    KeyErrorAmbiguous,
    find_or,
    typename,
)

Operand = Parameter | Expression


@dataclass(eq=True, frozen=True)
class ParamLink:
    operator: Expression
    operand: Operand


def _node(node: Operand):
    try:
        subtype = find_or(_GROUP_TYPES, lambda t: isinstance(node, t), default=Node)
    except KeyErrorAmbiguous as e:
        subtype = e.duplicates[0]

    hier = node.get_hierarchy()
    type_hier = [t for t, _ in hier]
    name_hier = [n for _, n in hier]
    name = ".".join(name_hier)
    types = "|".join(typename(t) for t in type_hier)
    label = f"{name}\n({types})"

    return {
        "data": {
            "id": str(id(node)),
            "label": label,
            "type": typename(subtype),
        }
    }


def _link(link: ParamLink):
    return {
        "data": {
            "source": str(id(link.operand)),
            "target": str(id(link.operator)),
        }
    }


class _Stylesheet:
    _BASE = [
        {
            "selector": "node",
            "style": {
                "content": "data(label)",
                "text-opacity": 0.8,
                "text-valign": "center",
                "text-halign": "center",
                "font-size": "0.3em",
                "background-color": "#BFD7B5",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 0.5,
                "border-width": 1,
                "border-color": "#888888",
                "border-opacity": 0.5,
                # group
                "font-weight": "bold",
                # "font-size": "1.5em",
                # "text-valign": "top",
                # "text-outline-color": "#FFFFFF",
                # "text-outline-width": 1.5,
                "text-wrap": "wrap",
                # "border-width": 4,
            },
        },
        {
            "selector": "edge",
            "style": {
                "width": 1,
                "line-color": "#A3C4BC",
                "curve-style": "bezier",
                "target-arrow-shape": "triangle",
                "arrow-scale": 1,
                "target-arrow-color": "#A3C4BC",
                "text-outline-color": "#FFFFFF",
                "text-outline-width": 2,
            },
        },
    ]

    def __init__(self):
        self.stylesheet = list(self._BASE)

    def add_node_type(self, node_type: str, color: str):
        self.stylesheet.append(
            {
                "selector": f'node[type = "{node_type}"]',
                "style": {"background-color": color},
            }
        )


DAGRE_LAYOUT = {
    # Dagre algorithm options (uses default value if undefined)
    "name": "dagre",
    # Separation between adjacent nodes in the same rank
    # "nodeSep": None,
    # Separation between adjacent edges in the same rank
    # "edgeSep": None,
    # Separation between each rank in the layout
    # "rankSep": None,
    # 'TB' for top to bottom flow, 'LR' for left to right
    # "rankDir": None,
    # Alignment for rank nodes. Can be 'UL', 'UR', 'DL', or 'DR'
    # "align": None,
    # If 'greedy', uses heuristic to find feedback arc set
    # "acyclicer": None,
    # Algorithm to assign rank to nodes: 'network-simplex', 'tight-tree'
    # or 'longest-path'
    # "ranker": "tight-tree",
    # Number of ranks to keep between source and target of the edge
    # "minLen": lambda edge: 1,
    # Higher weight edges are generally made shorter and straighter
    # "edgeWeight": lambda edge: 1,
    # General layout options
    # Whether to fit to viewport
    # "fit": True,
    # Fit padding
    # "padding": 30,
    # Factor to expand/compress overall area nodes take up
    # "spacingFactor": None,
    # Include labels in node space calculation
    # "nodeDimensionsIncludeLabels": False,
    # Whether to transition node positions
    # "animate": False,
    # Whether to animate specific nodes
    # "animateFilter": lambda node, i: True,
    # Duration of animation in ms if enabled
    # "animationDuration": 500,
    # Easing of animation if enabled
    # "animationEasing": None,
    # Constrain layout bounds: {x1, y1, x2, y2} or {x1, y1, w, h}
    # "boundingBox": None,
    # Function to transform final node position
    # "transform": lambda node, pos: pos,
    # Callback on layoutready
    # "ready": lambda: None,
    # Sorting function to order nodes and edges
    # "sort": None,
    # Callback on layoutstop
    # "stop": lambda: None,
}


def buttons(layout: Layout):
    app = layout.app

    layout_chooser = dcc.RadioItems(
        id="layout-radio",
        options=[
            {"label": "fcose", "value": "fcose"},
            {"label": "dagre", "value": "dagre"},
        ],
        value="dagre",
    )

    dagre_ranker = dcc.RadioItems(
        id="layout-dagre-ranker",
        options=[
            {"label": "network-simplex", "value": "network-simplex"},
            {"label": "tight-tree", "value": "tight-tree"},
            {"label": "longest-path", "value": "longest-path"},
        ],
        value="tight-tree",
    )

    html_controls = html.Div(
        className="controls",
        style={"padding": "10px", "background-color": "#f0f0f0"},
        children=[
            html.Table(
                [
                    html.Tr([html.Td("Layout:"), html.Td(layout_chooser)]),
                    html.Tr([html.Td("Dagre Ranker:"), html.Td(dagre_ranker)]),
                ]
            )
        ],
    )
    layout.div_children.insert(-2, html_controls)

    @app.callback(
        Output("graph-view", "layout"),
        Input("layout-radio", "value"),
        Input("layout-dagre-ranker", "value"),
        State("graph-view", "layout"),
    )
    def absolute_layout(layout_radio, layout_dagre_ranker, current_layout):
        # print(layout_radio, layout_dagre_ranker)
        layout.set_type(layout_radio, current_layout)

        if layout_dagre_ranker:
            current_layout["ranker"] = layout_dagre_ranker

        return current_layout


def visualize_parameters(G: Graph, height: int | None = None):
    Operand_ = (Parameter, Expression)
    nodes = GraphFunctions(G).nodes_of_types(Operand_)
    nodes = cast(list[Operand], nodes)

    edges = {
        ParamLink(n, e.node)
        for n in nodes
        if isinstance(n, Expression)
        for e, li in n.operates_on.edges.items()
        if not isinstance(li, LinkSibling)
        and e.node is not n
        and isinstance(e.node, Operand_)
    }

    # TODO filter equivalency classes

    parameter_report(G, Path("./build/params.txt"))

    elements = [_node(n) for n in nodes] + [_link(li) for li in edges]
    stylesheet = _Stylesheet()

    node_types_colors = [
        (typename(group_type), color) for group_type, color in _GROUP_TYPES.items()
    ]

    for node_type, color in node_types_colors:
        stylesheet.add_node_type(node_type, color)

    cyto.load_extra_layouts()
    app = Dash(__name__)
    app.layout = _Layout(stylesheet, elements, extra=DAGRE_LAYOUT)

    # Extra layouting
    layout = Layout(app, elements, list(nodes))
    buttons(layout)

    app.run(jupyter_height=height or 1400)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from abc import abstractmethod

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.trait import Trait

logger = logging.getLogger(__name__)


# TODO better name
class can_be_decoupled(Trait):
    @abstractmethod
    def decouple(
        self,
        owner: Module,
        count: int = 1,
    ) -> F.MultiCapacitor: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import time
import unittest
from itertools import combinations
from pathlib import Path

import pytest

from faebryk.libs.util import (
    DAG,
    SharedReference,
    assert_once,
    complete_type_string,
    invert_dict,
    list_match,
    once,
    path_replace,
    times_out,
    zip_non_locked,
)


class TestUtil(unittest.TestCase):
    def test_zip_non_locked(self):
        expected = [(1, 4), (2, 5), (3, 6)]

        for val, ref in zip(zip_non_locked([1, 2, 3], [4, 5, 6]), expected):
            self.assertEqual(val, ref)

        expected = [(1, 4), (2, 5), (2, 6)]
        for val, ref in zip((it := zip_non_locked([1, 2, 3], [4, 5, 6])), expected):
            self.assertEqual(val, ref)
            if val == (2, 5):
                it.advance(1)

    def test_shared_reference(self):
        def all_equal(*args: SharedReference):
            for left, right in combinations(args, 2):
                self.assertIs(left.links, right.links)
                self.assertIs(left(), right())
                self.assertEqual(left, right)

        r1 = SharedReference(1)
        r2 = SharedReference(2)

        self.assertEqual(r1(), 1)
        self.assertEqual(r2(), 2)

        r1.link(r2)

        all_equal(r1, r2)

        r3 = SharedReference(3)
        r3.link(r2)

        all_equal(r1, r2, r3)

        r4 = SharedReference(4)
        r5 = SharedReference(5)

        r4.link(r5)

        all_equal(r4, r5)

        r5.link(r1)

        all_equal(r1, r2, r3, r4, r5)

    def test_once(self):
        global ran
        ran = False

        @once
        def do(val: int):
            global ran
            ran = True
            return val

        self.assertFalse(ran)

        self.assertEqual(do(5), 5)
        self.assertTrue(ran)
        ran = False

        self.assertEqual(do(5), 5)
        self.assertFalse(ran)

        self.assertEqual(do(6), 6)
        self.assertTrue(ran)
        ran = False

        class A:
            @classmethod
            @once
            def do(cls):
                global ran
                ran = True
                return cls

            @once
            def do_inst(self, arg: int):
                global ran
                ran = True
                return arg

        self.assertEqual(A.do(), A)
        self.assertTrue(ran)
        ran = False

        self.assertEqual(A.do(), A)
        self.assertFalse(ran)

        ran = False
        a = A()
        self.assertEqual(a.do_inst(5), 5)
        self.assertTrue(ran)
        ran = False
        self.assertEqual(a.do_inst(5), 5)
        self.assertFalse(ran)

    def test_assert_once(self):
        class A:
            def __init__(self):
                self.a = 1

            @assert_once
            def do(self):
                print("do")
                self.a = 5

            @assert_once
            def do_with_arg(self, arg: int):
                print("do_with_arg", arg)
                self.a = arg

        a = A()
        a.do()
        self.assertEqual(a.a, 5)
        self.assertRaises(AssertionError, a.do)
        a.do_with_arg(3)
        self.assertEqual(a.a, 3)
        self.assertRaises(AssertionError, a.do_with_arg, 2)

    def test_times_out(self):
        @times_out(0.1)
        def do(wait_s: float):
            time.sleep(wait_s)

        self.assertRaises(TimeoutError, do, 0.5)
        do(0)


class _DictTestObjBrokenHash:
    def __init__(self, value: object):
        self.value = value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _DictTestObjBrokenHash):
            return False
        return self.value == other.value

    def __hash__(self) -> int:
        return 5


@pytest.mark.parametrize(
    "to_test, expected",
    [
        ({1: "a", 2: "b", 3: "c"}, {"a": [1], "b": [2], "c": [3]}),
        ({1: "a", 2: "a", 3: "a"}, {"a": [1, 2, 3]}),
        ({1: "a", 2: "b", 3: "a"}, {"a": [1, 3], "b": [2]}),
        (
            {
                1: (a := _DictTestObjBrokenHash("a")),
                2: (b := _DictTestObjBrokenHash("b")),
                3: a,
            },
            {a: [1, 3], b: [2]},
        ),
    ],
)
def test_invert_dict(to_test: dict, expected: dict):
    out = invert_dict(to_test)
    assert out == expected


def test_dag():
    dag = DAG[int]()
    dag.add_edge(1, 2)
    dag.add_edge(2, 3)

    assert dag.get(1).children == {2}
    assert dag.get(3).parents == {2}

    dag.add_edge(1, 3)
    assert dag.get(3).parents == {1, 2}
    assert dag.get(2).children == {3}
    assert dag.get(1).children == {2, 3}

    assert not dag.contains_cycles

    dag.add_edge(3, 1)
    assert dag.contains_cycles


def test_complete_type_string():
    a = {"a": 1, 5: object(), "c": {"a": 1}}
    assert complete_type_string(a) == "dict[str | int, int | object | dict[str, int]]"


def test_ordered_set():
    from ordered_set import OrderedSet

    s = OrderedSet([1, 4, 2, 3, 4])
    assert s == {1: None, 4: None, 2: None, 3: None}
    assert s == {1, 4, 2, 3}
    assert list(s) == [1, 4, 2, 3]

    x = s | {5, 7, 6}
    assert x == {1, 4, 2, 3, 5, 6, 7}
    assert list(x) == [1, 4, 2, 3, 5, 6, 7]


@pytest.mark.parametrize(
    "base, match, expected",
    [
        ([1, 2, 3], [2, 3], [1]),
        ([1, 2, 3], [2, 3, 4], []),
        ([1, 2, 3], [1, 2, 3], [0]),
        ([1, 2, 3], [1, 2, 3, 4], []),
        ([1, 2, 3], [1, 2, 3, 4, 5], []),
    ],
)
def test_list_match(base, match, expected):
    assert list(list_match(base, match)) == expected


@pytest.mark.parametrize(
    "base, match, replacement, expected",
    [
        (Path("a/b/c"), Path("b"), Path("d"), Path("a/d/c")),
        (Path("a/b/c"), Path("b/c"), Path("d"), Path("a/d")),
        (Path("a/c/c"), Path("b/c"), Path("d"), Path("a/c/c")),
        (Path("a/c/c"), Path("c"), Path("d"), Path("a/d/d")),
        (Path("a/c/c/c/c"), Path("c/c"), Path("d"), Path("a/d/d")),
    ],
)
def test_path_replace(base, match, replacement, expected):
    assert path_replace(base, match, replacement) == expected



#!/usr/bin/env python3

"""
This file generates faebryk/src/faebryk/library/__init__.py
"""

import logging
import re
from graphlib import TopologicalSorter
from itertools import groupby
from pathlib import Path
from typing import Iterable

logger = logging.getLogger(__name__)

REPO_ROOT = Path(__file__).parent.parent.parent
LIBRARY_DIR = REPO_ROOT / "src" / "faebryk" / "library"
OUT = LIBRARY_DIR / "_F.py"


def try_(stmt: str, exc: str | type[Exception] | Iterable[type[Exception]]):
    if isinstance(exc, type):
        exc = exc.__name__
    if not isinstance(exc, str):
        exc = f"({', '.join(e.__name__ for e in exc)})"

    return (
        f"try:\n    {stmt}\nexcept {exc} as e:\n    print('{stmt.split(' ')[-1]}', e)"
    )


def topo_sort(modules_out: dict[str, tuple[Path, str]]):
    def find_deps(module_path: Path) -> set[str]:
        f = module_path.read_text(encoding="utf-8")
        p = re.compile(r"[^a-zA-Z_0-9]F\.([a-zA-Z_][a-zA-Z_0-9]*)")
        return set(p.findall(f))

    if True:
        SRC_DIR = LIBRARY_DIR.parent.parent
        all_modules = [
            (p.stem, p) for p in SRC_DIR.rglob("*.py") if not p.stem.startswith("_")
        ]
    else:
        all_modules = [
            (module_name, module_path)
            for module_name, (module_path, _) in modules_out.items()
        ]

    topo_graph = [
        (module_name, find_deps(module_path))
        for module_name, module_path in all_modules
    ]

    # handles name collisions (of non-lib modules)
    topo_grouped = groupby(
        sorted(topo_graph, key=lambda item: item[0]), lambda item: item[0]
    )
    topo_merged = {
        name: {x for xs in group for x in xs[1]} for name, group in topo_grouped
    }

    # sort for deterministic order
    topo_graph = {
        k: sorted(v) for k, v in sorted(topo_merged.items(), key=lambda item: item[0])
    }
    order = list(TopologicalSorter(topo_graph).static_order())

    # TEST
    seen = set()
    for m in order:
        if m not in topo_graph:
            continue
        for sub in topo_graph[m]:
            if sub not in seen and sub in topo_graph:
                raise Exception(f"Collision: {sub} after {m}")
        seen.add(m)

    return [
        (module_name, modules_out[module_name][1])
        for module_name in order
        if module_name in modules_out
    ]


def main():
    assert LIBRARY_DIR.exists()

    logger.info(f"Scanning {LIBRARY_DIR} for modules")

    module_files = [p for p in LIBRARY_DIR.glob("*.py") if not p.name.startswith("_")]

    logger.info(f"Found {len(module_files)} modules")

    modules_out: dict[str, tuple[Path, str]] = {}

    # Import each module and add its class to the current namespace
    # for module_name in module_files:
    #    module = importlib.import_module(
    #        f"faebryk.library.{module_name}"  # , package=__name__
    #    )
    #    class_name = module_name
    #    if hasattr(module, class_name):
    #        # globals()[class_name] = getattr(module, class_name)
    #        modules_out[module_name] = class_name

    # assuming class name is equal to file stem
    modules_out = {
        module_path.stem: (module_path, module_path.stem)
        for module_path in module_files
    }

    modules_ordered = topo_sort(modules_out)

    logger.info(f"Found {len(modules_out)} classes")

    OUT.write_text(
        "# This file is part of the faebryk project\n"
        "# SPDX-License-Identifier: MIT\n"
        "\n"
        '"""\n'
        "This file is autogenerated by tools/library/gen_F.py\n"
        "This is the __init__.py file of the library\n"
        "All modules are in ./<module>.py with name class <module>\n"
        "Export all <module> classes here\n"
        "Do it programmatically instead of specializing each manually\n"
        "This way we can add new modules without changing this file\n"
        '"""\n'
        "\n"
        "# Disable ruff warning for whole block\n"
        "# flake8: noqa: F401\n"
        "# flake8: noqa: I001\n"
        "# flake8: noqa: E501\n"
        "\n"
        + "\n".join(
            # try_(
            f"from faebryk.library.{module} import {class_}"
            #    (AttributeError,),
            # )
            for module, class_ in modules_ordered
        )
        + "\n",
        encoding="utf-8",
    )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()
